import pytest
from pytest_django import DjangoAssertNumQueries
from strawberry.relay.utils import to_base64

from .schema import FruitModel, schema


@pytest.fixture(autouse=True)
def _fixtures(transactional_db):
    for pk, name, color in [
        (1, "Banana", "yellow"),
        (2, "Apple", "red"),
        (3, "Pineapple", "yellow"),
        (4, "Grape", "purple"),
        (5, "Orange", "orange"),
    ]:
        FruitModel.objects.create(
            id=pk,
            name=name,
            color=color,
        )


def test_query_node():
    result = schema.execute_sync(
        """
        query TestQuery ($id: ID!) {
            node (id: $id) {
                ... on Node {
                    id
                }
                ... on Fruit {
                    name
                    color
                }
            }
        }
        """,
        variable_values={
            "id": to_base64("Fruit", 2),
        },
    )
    assert result.errors is None
    assert result.data == {
        "node": {
            "id": to_base64("Fruit", 2),
            "color": "red",
            "name": "Apple",
        },
    }


async def test_query_node_with_async_permissions():
    result = await schema.execute(
        """
        query TestQuery ($id: ID!) {
            nodeWithAsyncPermissions (id: $id) {
                ... on Node {
                    id
                }
                ... on Fruit {
                    name
                    color
                }
            }
        }
        """,
        variable_values={
            "id": to_base64("Fruit", 2),
        },
    )
    assert result.errors is None
    assert result.data == {
        "nodeWithAsyncPermissions": {
            "id": to_base64("Fruit", 2),
            "color": "red",
            "name": "Apple",
        },
    }


def test_query_node_optional():
    result = schema.execute_sync(
        """
        query TestQuery ($id: ID!) {
            nodeOptional (id: $id) {
                ... on Node {
                    id
                }
                ... on Fruit {
                    name
                    color
                }
            }
        }
        """,
        variable_values={
            "id": to_base64("Fruit", 999),
        },
    )
    assert result.errors is None
    assert result.data == {"nodeOptional": None}


async def test_query_node_async():
    result = await schema.execute(
        """
        query TestQuery ($id: ID!) {
            node (id: $id) {
                ... on Node {
                    id
                }
                ... on Fruit {
                    name
                    color
                }
            }
        }
        """,
        variable_values={
            "id": to_base64("Fruit", 2),
        },
    )
    assert result.errors is None
    assert result.data == {
        "node": {
            "id": to_base64("Fruit", 2),
            "color": "red",
            "name": "Apple",
        },
    }


async def test_query_node_optional_async():
    result = await schema.execute(
        """
        query TestQuery ($id: ID!) {
            nodeOptional (id: $id) {
                ... on Node {
                    id
                }
                ... on Fruit {
                    name
                    color
                }
            }
        }
        """,
        variable_values={
            "id": to_base64("Fruit", 999),
        },
    )
    assert result.errors is None
    assert result.data == {"nodeOptional": None}


def test_query_nodes():
    result = schema.execute_sync(
        """
        query TestQuery ($ids: [ID!]!) {
            nodes (ids: $ids) {
                ... on Node {
                    id
                }
                ... on Fruit {
                    name
                    color
                }
            }
        }
        """,
        variable_values={
            "ids": [to_base64("Fruit", 2), to_base64("Fruit", 4)],
        },
    )
    assert result.errors is None
    assert result.data == {
        "nodes": [
            {
                "id": to_base64("Fruit", 2),
                "name": "Apple",
                "color": "red",
            },
            {
                "id": to_base64("Fruit", 4),
                "name": "Grape",
                "color": "purple",
            },
        ],
    }


def test_query_nodes_optional():
    result = schema.execute_sync(
        """
        query TestQuery ($ids: [ID!]!) {
            nodesOptional (ids: $ids) {
                ... on Node {
                    id
                }
                ... on Fruit {
                    name
                    color
                }
            }
        }
        """,
        variable_values={
            "ids": [
                to_base64("Fruit", 2),
                to_base64("Fruit", 999),
                to_base64("Fruit", 4),
            ],
        },
    )
    assert result.errors is None
    assert result.data == {
        "nodesOptional": [
            {
                "id": to_base64("Fruit", 2),
                "name": "Apple",
                "color": "red",
            },
            None,
            {
                "id": to_base64("Fruit", 4),
                "name": "Grape",
                "color": "purple",
            },
        ],
    }


async def test_query_nodes_async():
    result = await schema.execute(
        """
        query TestQuery ($ids: [ID!]!) {
            nodes (ids: $ids) {
                ... on Node {
                    id
                }
                ... on Fruit {
                    name
                    color
                }
            }
        }
        """,
        variable_values={
            "ids": [
                to_base64("Fruit", 2),
                to_base64("Fruit", 4),
            ],
        },
    )
    assert result.errors is None
    assert result.data == {
        "nodes": [
            {
                "id": to_base64("Fruit", 2),
                "name": "Apple",
                "color": "red",
            },
            {
                "id": to_base64("Fruit", 4),
                "name": "Grape",
                "color": "purple",
            },
        ],
    }


async def test_query_nodes_optional_async():
    result = await schema.execute(
        """
        query TestQuery ($ids: [ID!]!) {
            nodesOptional (ids: $ids) {
                ... on Node {
                    id
                }
                ... on Fruit {
                    name
                    color
                }
            }
        }
        """,
        variable_values={
            "ids": [
                to_base64("Fruit", 2),
                to_base64("Fruit", 998),
                to_base64("Fruit", 4),
                to_base64("Fruit", 999),
            ],
        },
    )
    assert result.errors is None
    assert result.data == {
        "nodesOptional": [
            {
                "id": to_base64("Fruit", 2),
                "name": "Apple",
                "color": "red",
            },
            None,
            {
                "id": to_base64("Fruit", 4),
                "name": "Grape",
                "color": "purple",
            },
            None,
        ],
    }


fruits_query = """
query TestQuery (
    $first: Int = null
    $last: Int = null
    $before: String = null,
    $after: String = null,
) {{
    {} (
        first: $first
        last: $last
        before: $before
        after: $after
    ) {{
        pageInfo {{
            hasNextPage
            hasPreviousPage
            startCursor
            endCursor
        }}
        edges {{
            cursor
            node {{
                id
                name
                color
            }}
        }}
    }}
}}
"""

attrs = [
    "fruits",
    "fruitsLazy",
    "fruitsWithFiltersAndOrder",
    "fruitsCustomResolver",
    "fruitsCustomResolverWithFiltersAndOrder",
]


@pytest.mark.parametrize("query_attr", attrs)
def test_query_connection(query_attr: str):
    result = schema.execute_sync(
        fruits_query.format(query_attr),
        variable_values={},
    )
    assert result.errors is None
    assert result.data == {
        query_attr: {
            "edges": [
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjA=",
                    "node": {
                        "id": to_base64("Fruit", 1),
                        "color": "yellow",
                        "name": "Banana",
                    },
                },
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjE=",
                    "node": {
                        "id": to_base64("Fruit", 2),
                        "color": "red",
                        "name": "Apple",
                    },
                },
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjI=",
                    "node": {
                        "id": to_base64("Fruit", 3),
                        "color": "yellow",
                        "name": "Pineapple",
                    },
                },
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjM=",
                    "node": {
                        "id": to_base64("Fruit", 4),
                        "color": "purple",
                        "name": "Grape",
                    },
                },
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjQ=",
                    "node": {
                        "id": to_base64("Fruit", 5),
                        "color": "orange",
                        "name": "Orange",
                    },
                },
            ],
            "pageInfo": {
                "hasNextPage": False,
                "hasPreviousPage": False,
                "startCursor": to_base64("arrayconnection", "0"),
                "endCursor": to_base64("arrayconnection", "4"),
            },
        },
    }


@pytest.mark.parametrize("query_attr", attrs)
async def test_query_connection_async(query_attr: str):
    result = await schema.execute(
        fruits_query.format(query_attr),
        variable_values={},
    )
    assert result.errors is None
    assert result.data == {
        query_attr: {
            "edges": [
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjA=",
                    "node": {
                        "id": to_base64("Fruit", 1),
                        "color": "yellow",
                        "name": "Banana",
                    },
                },
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjE=",
                    "node": {
                        "id": to_base64("Fruit", 2),
                        "color": "red",
                        "name": "Apple",
                    },
                },
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjI=",
                    "node": {
                        "id": to_base64("Fruit", 3),
                        "color": "yellow",
                        "name": "Pineapple",
                    },
                },
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjM=",
                    "node": {
                        "id": to_base64("Fruit", 4),
                        "color": "purple",
                        "name": "Grape",
                    },
                },
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjQ=",
                    "node": {
                        "id": to_base64("Fruit", 5),
                        "color": "orange",
                        "name": "Orange",
                    },
                },
            ],
            "pageInfo": {
                "hasNextPage": False,
                "hasPreviousPage": False,
                "startCursor": to_base64("arrayconnection", "0"),
                "endCursor": to_base64("arrayconnection", "4"),
            },
        },
    }


@pytest.mark.parametrize("query_attr", attrs)
def test_query_connection_filtering_first(query_attr: str):
    result = schema.execute_sync(
        fruits_query.format(query_attr),
        variable_values={"first": 2},
    )
    assert result.errors is None
    assert result.data == {
        query_attr: {
            "edges": [
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjA=",
                    "node": {
                        "id": to_base64("Fruit", 1),
                        "color": "yellow",
                        "name": "Banana",
                    },
                },
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjE=",
                    "node": {
                        "id": to_base64("Fruit", 2),
                        "color": "red",
                        "name": "Apple",
                    },
                },
            ],
            "pageInfo": {
                "hasNextPage": True,
                "hasPreviousPage": False,
                "startCursor": to_base64("arrayconnection", "0"),
                "endCursor": to_base64("arrayconnection", "1"),
            },
        },
    }


@pytest.mark.parametrize("query_attr", attrs)
async def test_query_connection_filtering_first_async(query_attr: str):
    result = await schema.execute(
        fruits_query.format(query_attr),
        variable_values={"first": 2},
    )
    assert result.errors is None
    assert result.data == {
        query_attr: {
            "edges": [
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjA=",
                    "node": {
                        "id": to_base64("Fruit", 1),
                        "color": "yellow",
                        "name": "Banana",
                    },
                },
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjE=",
                    "node": {
                        "id": to_base64("Fruit", 2),
                        "color": "red",
                        "name": "Apple",
                    },
                },
            ],
            "pageInfo": {
                "hasNextPage": True,
                "hasPreviousPage": False,
                "startCursor": to_base64("arrayconnection", "0"),
                "endCursor": to_base64("arrayconnection", "1"),
            },
        },
    }


@pytest.mark.parametrize("query_attr", attrs)
def test_query_connection_filtering_first_with_after(query_attr: str):
    result = schema.execute_sync(
        fruits_query.format(query_attr),
        variable_values={"first": 2, "after": to_base64("arrayconnection", "1")},
    )
    assert result.errors is None
    assert result.data == {
        query_attr: {
            "edges": [
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjI=",
                    "node": {
                        "id": to_base64("Fruit", 3),
                        "color": "yellow",
                        "name": "Pineapple",
                    },
                },
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjM=",
                    "node": {
                        "id": to_base64("Fruit", 4),
                        "color": "purple",
                        "name": "Grape",
                    },
                },
            ],
            "pageInfo": {
                "hasNextPage": True,
                "hasPreviousPage": True,
                "startCursor": to_base64("arrayconnection", "2"),
                "endCursor": to_base64("arrayconnection", "3"),
            },
        },
    }


@pytest.mark.parametrize("query_attr", attrs)
async def test_query_connection_filtering_first_with_after_async(query_attr: str):
    result = await schema.execute(
        fruits_query.format(query_attr),
        variable_values={"first": 2, "after": to_base64("arrayconnection", "1")},
    )
    assert result.errors is None
    assert result.data == {
        query_attr: {
            "edges": [
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjI=",
                    "node": {
                        "id": to_base64("Fruit", 3),
                        "color": "yellow",
                        "name": "Pineapple",
                    },
                },
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjM=",
                    "node": {
                        "id": to_base64("Fruit", 4),
                        "color": "purple",
                        "name": "Grape",
                    },
                },
            ],
            "pageInfo": {
                "hasNextPage": True,
                "hasPreviousPage": True,
                "startCursor": to_base64("arrayconnection", "2"),
                "endCursor": to_base64("arrayconnection", "3"),
            },
        },
    }


@pytest.mark.parametrize("query_attr", attrs)
def test_query_connection_filtering_last(query_attr: str):
    result = schema.execute_sync(
        fruits_query.format(query_attr),
        variable_values={"last": 2},
    )
    assert result.errors is None
    assert result.data == {
        query_attr: {
            "edges": [
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjM=",
                    "node": {
                        "id": to_base64("Fruit", 4),
                        "color": "purple",
                        "name": "Grape",
                    },
                },
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjQ=",
                    "node": {
                        "id": to_base64("Fruit", 5),
                        "color": "orange",
                        "name": "Orange",
                    },
                },
            ],
            "pageInfo": {
                "hasNextPage": False,
                "hasPreviousPage": True,
                "startCursor": to_base64("arrayconnection", "3"),
                "endCursor": to_base64("arrayconnection", "4"),
            },
        },
    }


@pytest.mark.parametrize("query_attr", attrs)
async def test_query_connection_filtering_last_async(query_attr: str):
    result = await schema.execute(
        fruits_query.format(query_attr),
        variable_values={"last": 2},
    )
    assert result.errors is None
    assert result.data == {
        query_attr: {
            "edges": [
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjM=",
                    "node": {
                        "id": to_base64("Fruit", 4),
                        "color": "purple",
                        "name": "Grape",
                    },
                },
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjQ=",
                    "node": {
                        "id": to_base64("Fruit", 5),
                        "color": "orange",
                        "name": "Orange",
                    },
                },
            ],
            "pageInfo": {
                "hasNextPage": False,
                "hasPreviousPage": True,
                "startCursor": to_base64("arrayconnection", "3"),
                "endCursor": to_base64("arrayconnection", "4"),
            },
        },
    }


@pytest.mark.parametrize("query_attr", attrs)
def test_query_connection_filtering_first_with_before(query_attr: str):
    result = schema.execute_sync(
        fruits_query.format(query_attr),
        variable_values={"first": 1, "before": to_base64("arrayconnection", "3")},
    )
    assert result.errors is None
    assert result.data == {
        query_attr: {
            "edges": [
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjI=",
                    "node": {
                        "id": to_base64("Fruit", 3),
                        "color": "yellow",
                        "name": "Pineapple",
                    },
                },
            ],
            "pageInfo": {
                "hasNextPage": True,
                "hasPreviousPage": True,
                "startCursor": to_base64("arrayconnection", "2"),
                "endCursor": to_base64("arrayconnection", "2"),
            },
        },
    }


@pytest.mark.parametrize("query_attr", attrs)
async def test_query_connection_filtering_first_with_before_async(query_attr: str):
    result = await schema.execute(
        fruits_query.format(query_attr),
        variable_values={"first": 1, "before": to_base64("arrayconnection", "3")},
    )
    assert result.errors is None
    assert result.data == {
        query_attr: {
            "edges": [
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjI=",
                    "node": {
                        "id": to_base64("Fruit", 3),
                        "color": "yellow",
                        "name": "Pineapple",
                    },
                },
            ],
            "pageInfo": {
                "hasNextPage": True,
                "hasPreviousPage": True,
                "startCursor": to_base64("arrayconnection", "2"),
                "endCursor": to_base64("arrayconnection", "2"),
            },
        },
    }


@pytest.mark.parametrize("query_attr", attrs)
def test_query_connection_filtering_last_with_before(query_attr: str):
    result = schema.execute_sync(
        fruits_query.format(query_attr),
        variable_values={"last": 2, "before": to_base64("arrayconnection", "4")},
    )
    assert result.errors is None
    assert result.data == {
        query_attr: {
            "edges": [
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjI=",
                    "node": {
                        "id": to_base64("Fruit", 3),
                        "color": "yellow",
                        "name": "Pineapple",
                    },
                },
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjM=",
                    "node": {
                        "id": to_base64("Fruit", 4),
                        "color": "purple",
                        "name": "Grape",
                    },
                },
            ],
            "pageInfo": {
                "hasNextPage": True,
                "hasPreviousPage": True,
                "startCursor": to_base64("arrayconnection", "2"),
                "endCursor": to_base64("arrayconnection", "3"),
            },
        },
    }


@pytest.mark.parametrize("query_attr", attrs)
async def test_query_connection_filtering_last_with_before_async(query_attr: str):
    result = await schema.execute(
        fruits_query.format(query_attr),
        variable_values={"last": 2, "before": to_base64("arrayconnection", "4")},
    )
    assert result.errors is None
    assert result.data == {
        query_attr: {
            "edges": [
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjI=",
                    "node": {
                        "id": to_base64("Fruit", 3),
                        "color": "yellow",
                        "name": "Pineapple",
                    },
                },
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjM=",
                    "node": {
                        "id": to_base64("Fruit", 4),
                        "color": "purple",
                        "name": "Grape",
                    },
                },
            ],
            "pageInfo": {
                "hasNextPage": True,
                "hasPreviousPage": True,
                "startCursor": to_base64("arrayconnection", "2"),
                "endCursor": to_base64("arrayconnection", "3"),
            },
        },
    }


fruits_query_filters_order = """
query TestQuery (
    $first: Int = null
    $last: Int = null
    $before: String = null,
    $after: String = null,
    $filters: FruitFilter
    $order: FruitOrder
) {{
    {} (
        first: $first
        last: $last
        before: $before
        after: $after
        filters: $filters
        order: $order
    ) {{
        pageInfo {{
            hasNextPage
            hasPreviousPage
            startCursor
            endCursor
        }}
        edges {{
            cursor
            node {{
                id
                name
                color
            }}
        }}
    }}
}}
"""

custom_attrs = [
    "fruitsWithFiltersAndOrder",
    "fruitsCustomResolverWithFiltersAndOrder",
]


@pytest.mark.parametrize("query_attr", custom_attrs)
def test_query_connection_with_filters(query_attr: str):
    result = schema.execute_sync(
        fruits_query_filters_order.format(query_attr),
        variable_values={"filters": {"name": {"endsWith": "e"}}},
    )
    assert result.errors is None
    assert result.data == {
        query_attr: {
            "edges": [
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjA=",
                    "node": {
                        "id": to_base64("Fruit", 2),
                        "color": "red",
                        "name": "Apple",
                    },
                },
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjE=",
                    "node": {
                        "id": to_base64("Fruit", 3),
                        "color": "yellow",
                        "name": "Pineapple",
                    },
                },
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjI=",
                    "node": {
                        "id": to_base64("Fruit", 4),
                        "color": "purple",
                        "name": "Grape",
                    },
                },
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjM=",
                    "node": {
                        "id": to_base64("Fruit", 5),
                        "color": "orange",
                        "name": "Orange",
                    },
                },
            ],
            "pageInfo": {
                "hasNextPage": False,
                "hasPreviousPage": False,
                "startCursor": to_base64("arrayconnection", "0"),
                "endCursor": to_base64("arrayconnection", "3"),
            },
        },
    }


@pytest.mark.parametrize("query_attr", custom_attrs)
def test_query_connection_with_filters_and_order(query_attr: str):
    result = schema.execute_sync(
        fruits_query_filters_order.format(query_attr),
        variable_values={
            "filters": {"name": {"endsWith": "e"}},
            "order": {"name": "DESC"},
        },
    )
    assert result.errors is None
    assert result.data == {
        query_attr: {
            "edges": [
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjA=",
                    "node": {
                        "id": to_base64("Fruit", 3),
                        "color": "yellow",
                        "name": "Pineapple",
                    },
                },
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjE=",
                    "node": {
                        "id": to_base64("Fruit", 5),
                        "color": "orange",
                        "name": "Orange",
                    },
                },
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjI=",
                    "node": {
                        "id": to_base64("Fruit", 4),
                        "color": "purple",
                        "name": "Grape",
                    },
                },
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjM=",
                    "node": {
                        "id": to_base64("Fruit", 2),
                        "color": "red",
                        "name": "Apple",
                    },
                },
            ],
            "pageInfo": {
                "hasNextPage": False,
                "hasPreviousPage": False,
                "startCursor": to_base64("arrayconnection", "0"),
                "endCursor": to_base64("arrayconnection", "3"),
            },
        },
    }


@pytest.mark.parametrize("query_attr", custom_attrs)
def test_query_connection_custom_resolver_filtering_first(query_attr: str):
    result = schema.execute_sync(
        fruits_query_filters_order.format(query_attr),
        variable_values={"first": 2, "filters": {"name": {"endsWith": "e"}}},
    )
    assert result.errors is None
    assert result.data == {
        query_attr: {
            "edges": [
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjA=",
                    "node": {
                        "id": to_base64("Fruit", 2),
                        "color": "red",
                        "name": "Apple",
                    },
                },
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjE=",
                    "node": {
                        "id": to_base64("Fruit", 3),
                        "color": "yellow",
                        "name": "Pineapple",
                    },
                },
            ],
            "pageInfo": {
                "hasNextPage": True,
                "hasPreviousPage": False,
                "startCursor": to_base64("arrayconnection", "0"),
                "endCursor": to_base64("arrayconnection", "1"),
            },
        },
    }


@pytest.mark.parametrize("query_attr", custom_attrs)
def test_query_connection_custom_resolver_filtering_first_with_after(query_attr: str):
    result = schema.execute_sync(
        fruits_query_filters_order.format(query_attr),
        variable_values={
            "first": 2,
            "after": to_base64("arrayconnection", "1"),
            "filters": {"name": {"endsWith": "e"}},
        },
    )
    assert result.errors is None
    assert result.data == {
        query_attr: {
            "edges": [
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjI=",
                    "node": {
                        "id": to_base64("Fruit", 4),
                        "color": "purple",
                        "name": "Grape",
                    },
                },
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjM=",
                    "node": {
                        "id": to_base64("Fruit", 5),
                        "color": "orange",
                        "name": "Orange",
                    },
                },
            ],
            "pageInfo": {
                "hasNextPage": False,
                "hasPreviousPage": True,
                "startCursor": to_base64("arrayconnection", "2"),
                "endCursor": to_base64("arrayconnection", "3"),
            },
        },
    }


@pytest.mark.parametrize("query_attr", custom_attrs)
def test_query_connection_custom_resolver_filtering_last(query_attr: str):
    result = schema.execute_sync(
        fruits_query_filters_order.format(query_attr),
        variable_values={"last": 2, "filters": {"name": {"endsWith": "e"}}},
    )
    assert result.errors is None
    assert result.data == {
        query_attr: {
            "edges": [
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjI=",
                    "node": {
                        "id": to_base64("Fruit", 4),
                        "color": "purple",
                        "name": "Grape",
                    },
                },
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjM=",
                    "node": {
                        "id": to_base64("Fruit", 5),
                        "color": "orange",
                        "name": "Orange",
                    },
                },
            ],
            "pageInfo": {
                "hasNextPage": False,
                "hasPreviousPage": True,
                "startCursor": to_base64("arrayconnection", "2"),
                "endCursor": to_base64("arrayconnection", "3"),
            },
        },
    }


@pytest.mark.parametrize("query_attr", custom_attrs)
def test_query_connection_custom_resolver_filtering_last_with_before(query_attr: str):
    result = schema.execute_sync(
        fruits_query_filters_order.format(query_attr),
        variable_values={
            "last": 2,
            "before": to_base64("arrayconnection", "3"),
            "filters": {"name": {"endsWith": "e"}},
        },
    )
    assert result.errors is None
    assert result.data == {
        query_attr: {
            "edges": [
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjE=",
                    "node": {
                        "id": to_base64("Fruit", 3),
                        "color": "yellow",
                        "name": "Pineapple",
                    },
                },
                {
                    "cursor": "YXJyYXljb25uZWN0aW9uOjI=",
                    "node": {
                        "id": to_base64("Fruit", 4),
                        "color": "purple",
                        "name": "Grape",
                    },
                },
            ],
            "pageInfo": {
                "hasNextPage": True,
                "hasPreviousPage": True,
                "startCursor": to_base64("arrayconnection", "1"),
                "endCursor": to_base64("arrayconnection", "2"),
            },
        },
    }


fruits_query_total_count = """
query TestQuery (
    $first: Int = null
    $last: Int = null
    $before: String = null,
    $after: String = null,
) {{
    {} (
        first: $first
        last: $last
        before: $before
        after: $after
    ) {{
        totalCount
    }}
}}
"""

attrs = [
    "fruits",
    "fruitsLazy",
    "fruitsWithFiltersAndOrder",
    "fruitsCustomResolver",
    "fruitsCustomResolverWithFiltersAndOrder",
]


@pytest.mark.parametrize("query_attr", custom_attrs)
def test_query_connection_total_count_sql_queries(
    django_assert_num_queries: DjangoAssertNumQueries, query_attr: str
):
    with django_assert_num_queries(1):
        result = schema.execute_sync(
            fruits_query_total_count.format(query_attr),
            variable_values={},
        )
    assert result.errors is None
    assert result.data == {
        query_attr: {"totalCount": 5},
    }



# ruff: noqa: TRY002, B904, BLE001, F811, PT012
import warnings
from typing import Any, cast
from unittest import mock

import pytest
import strawberry
from django.db.models import Case, Count, Value, When
from pytest_mock import MockFixture
from strawberry import Info, auto
from strawberry.annotation import StrawberryAnnotation
from strawberry.exceptions import MissingArgumentsAnnotationsError
from strawberry.types import get_object_definition
from strawberry.types.base import (
    StrawberryOptional,
    WithStrawberryObjectDefinition,
    get_object_definition,
)
from strawberry.types.field import StrawberryField

import strawberry_django
from strawberry_django.exceptions import (
    ForbiddenFieldArgumentError,
    MissingFieldArgumentError,
)
from strawberry_django.fields.field import StrawberryDjangoField
from strawberry_django.fields.filter_order import (
    FilterOrderField,
    FilterOrderFieldResolver,
)
from strawberry_django.ordering import Ordering, OrderSequence, process_order
from tests import models, utils
from tests.types import Fruit


@strawberry_django.ordering.order(models.Color)
class ColorOrder:
    pk: auto

    @strawberry_django.order_field
    def name(self, prefix, value: auto):
        return [value.resolve(f"{prefix}name")]


@strawberry_django.order_type(models.Fruit)
class FruitOrdering:
    name: auto


@strawberry_django.ordering.order(models.Fruit)
class FruitOrder:
    color_id: auto
    name: auto
    sweetness: auto
    color: "ColorOrder | None"

    @strawberry_django.order_field
    def types_number(self, queryset, prefix, value: auto):
        return queryset.annotate(
            count=Count(f"{prefix}types__id"),
            count_nulls=Case(
                When(count=0, then=Value(None)),
                default="count",
            ),
        ), [value.resolve("count_nulls")]


@strawberry_django.type(models.Fruit, order=FruitOrder)
class FruitWithOrder:
    id: auto
    name: auto


@strawberry.type
class Query:
    fruits: list[Fruit] = strawberry_django.field(order=FruitOrder)
    fruits_with_ordering: list[Fruit] = strawberry_django.field(
        order=FruitOrder, ordering=FruitOrdering
    )


@pytest.fixture
def query():
    return utils.generate_query(Query)


def test_legacy_order_argument_is_deprecated():
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always", DeprecationWarning)
        strawberry_django.field(order=FruitOrder)
        assert len(w) == 1
        assert issubclass(w[-1].category, DeprecationWarning)
        assert (
            str(w[-1].message)
            == "strawberry_django.order is deprecated in favor of strawberry_django.order_type."
        )


def test_legacy_order_type_is_deprecated():
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always", DeprecationWarning)

        @strawberry_django.ordering.order(models.Fruit)
        class TestOrder:
            name: auto

        assert len(w) == 1
        assert issubclass(w[-1].category, DeprecationWarning)
        assert (
            str(w[-1].message)
            == "strawberry_django.order is deprecated in favor of strawberry_django.order_type."
        )


def test_legacy_order_works_when_ordering_is_present(query, fruits):
    result = query("{ fruitsWithOrdering(order: { name: ASC }) { id name } }")
    assert not result.errors
    assert result.data["fruitsWithOrdering"] == [
        {"id": "3", "name": "banana"},
        {"id": "2", "name": "raspberry"},
        {"id": "1", "name": "strawberry"},
    ]


def test_ordering_works_when_legacy_order_is_present(query, fruits):
    result = query("{ fruitsWithOrdering(ordering: [{ name: ASC }]) { id name } }")
    assert not result.errors
    assert result.data["fruitsWithOrdering"] == [
        {"id": "3", "name": "banana"},
        {"id": "2", "name": "raspberry"},
        {"id": "1", "name": "strawberry"},
    ]


def test_error_when_ordering_and_order_given(query, fruits):
    result = query(
        "{ fruitsWithOrdering(ordering: [{ name: ASC }], order: { name: ASC }) { id name } }"
    )
    assert result.errors is not None
    assert len(result.errors) == 1
    assert (
        result.errors[0].message == "Only one of `ordering` or `order` must be given."
    )


def test_field_order_definition():
    field = StrawberryDjangoField(type_annotation=StrawberryAnnotation(FruitWithOrder))
    assert field.get_order() == FruitOrder
    field = StrawberryDjangoField(
        type_annotation=StrawberryAnnotation(FruitWithOrder),
        filters=None,
    )
    assert field.get_filters() is None


def test_asc(query, fruits):
    result = query("{ fruits(order: { name: ASC }) { id name } }")
    assert not result.errors
    assert result.data["fruits"] == [
        {"id": "3", "name": "banana"},
        {"id": "2", "name": "raspberry"},
        {"id": "1", "name": "strawberry"},
    ]


def test_desc(query, fruits):
    result = query("{ fruits(order: { name: DESC }) { id name } }")
    assert not result.errors
    assert result.data["fruits"] == [
        {"id": "1", "name": "strawberry"},
        {"id": "2", "name": "raspberry"},
        {"id": "3", "name": "banana"},
    ]


def test_relationship(query, fruits):
    def add_color(fruit, color_name):
        fruit.color = models.Color.objects.create(name=color_name)
        fruit.save()

    color_names = ["red", "dark red", "yellow"]
    for fruit, color_name in zip(fruits, color_names, strict=False):
        add_color(fruit, color_name)
    result = query(
        "{ fruits(order: { color: { name: DESC } }) { id name color { name } } }",
    )
    assert not result.errors
    assert result.data["fruits"] == [
        {"id": "3", "name": "banana", "color": {"name": "yellow"}},
        {"id": "1", "name": "strawberry", "color": {"name": "red"}},
        {"id": "2", "name": "raspberry", "color": {"name": "dark red"}},
    ]


def test_arguments_order_respected(query, db):
    yellow = models.Color.objects.create(name="yellow")
    red = models.Color.objects.create(name="red")

    f1 = models.Fruit.objects.create(
        name="strawberry",
        sweetness=1,
        color=red,
    )
    f2 = models.Fruit.objects.create(
        name="banana",
        sweetness=2,
        color=yellow,
    )
    f3 = models.Fruit.objects.create(
        name="apple",
        sweetness=0,
        color=red,
    )

    result = query("{ fruits(order: { name: ASC, sweetness: ASC }) { id } }")
    assert not result.errors
    assert result.data["fruits"] == [{"id": str(f.pk)} for f in [f3, f2, f1]]

    result = query("{ fruits(order: { sweetness: DESC, name: ASC }) { id } }")
    assert not result.errors
    assert result.data["fruits"] == [{"id": str(f.pk)} for f in [f2, f1, f3]]

    result = query("{ fruits(order: { color: {name: ASC}, name: ASC }) { id } }")
    assert not result.errors
    assert result.data["fruits"] == [{"id": str(f.pk)} for f in [f3, f1, f2]]

    result = query("{ fruits(order: { color: {pk: ASC}, name: ASC }) { id } }")
    assert not result.errors
    assert result.data["fruits"] == [{"id": str(f.pk)} for f in [f2, f3, f1]]

    result = query("{ fruits(order: { colorId: ASC, name: ASC }) { id } }")
    assert not result.errors
    assert result.data["fruits"] == [{"id": str(f.pk)} for f in [f2, f3, f1]]

    result = query("{ fruits(order: { name: ASC, colorId: ASC }) { id } }")
    assert not result.errors
    assert result.data["fruits"] == [{"id": str(f.pk)} for f in [f3, f2, f1]]


def test_order_sequence():
    f1 = StrawberryField(graphql_name="sOmEnAmE", python_name="some_name")
    f2 = StrawberryField(python_name="some_name")

    assert OrderSequence.get_graphql_name(None, f1) == "sOmEnAmE"
    assert OrderSequence.get_graphql_name(None, f2) == "someName"

    assert OrderSequence.sorted(None, None, fields=[f1, f2]) == [f1, f2]

    sequence = {"someName": OrderSequence(0, None), "sOmEnAmE": OrderSequence(1, None)}
    assert OrderSequence.sorted(None, sequence, fields=[f1, f2]) == [f1, f2]


def test_order_type():
    @strawberry_django.ordering.order(models.Fruit)
    class FruitOrder:
        color_id: auto
        name: auto
        sweetness: auto

        @strawberry_django.order_field
        def custom_order(self, value: auto, prefix: str):
            pass

    annotated_type = StrawberryOptional(Ordering.__strawberry_definition__)  # type: ignore

    assert [
        (
            f.name,
            f.__class__,
            f.type,
            f.base_resolver.__class__ if f.base_resolver else None,
        )
        for f in get_object_definition(FruitOrder, strict=True).fields
    ] == [
        ("color_id", StrawberryField, annotated_type, None),
        ("name", StrawberryField, annotated_type, None),
        ("sweetness", StrawberryField, annotated_type, None),
        (
            "custom_order",
            FilterOrderField,
            annotated_type,
            FilterOrderFieldResolver,
        ),
    ]


def test_order_field_missing_prefix():
    with pytest.raises(
        MissingFieldArgumentError, match=r".*\"prefix\".*\"field_method\".*"
    ):

        @strawberry_django.order_field
        def field_method():
            pass


def test_order_field_missing_value():
    with pytest.raises(
        MissingFieldArgumentError, match=r".*\"value\".*\"field_method\".*"
    ):

        @strawberry_django.order_field
        def field_method(prefix):
            pass


def test_order_field_missing_value_annotation():
    with pytest.raises(
        MissingArgumentsAnnotationsError,
        match=r"Missing annotation.*\"value\".*\"field_method\".*",
    ):

        @strawberry_django.order_field
        def field_method(prefix, value):
            pass


def test_order_field():
    try:

        @strawberry_django.order_field
        def field_method(
            self, root, info: Info, prefix, value: auto, sequence, queryset
        ):
            pass

    except Exception as exc:
        raise pytest.fail(f"DID RAISE {exc}")


def test_order_field_forbidden_param_annotation():
    with pytest.raises(
        MissingArgumentsAnnotationsError,
        match=r".*\"forbidden_param\".*\"field_method\".*",
    ):

        @strawberry_django.order_field
        def field_method(prefix, value: auto, sequence, queryset, forbidden_param):
            pass


def test_order_field_forbidden_param():
    with pytest.raises(
        ForbiddenFieldArgumentError,
        match=r".*\"forbidden_param\".*\"field_method\".*",
    ):

        @strawberry_django.order_field
        def field_method(prefix, value: auto, sequence, queryset, forbidden_param: str):
            pass


def test_order_field_missing_queryset():
    with pytest.raises(MissingFieldArgumentError, match=r".*\"queryset\".*\"order\".*"):

        @strawberry_django.order_field
        def order(prefix):
            pass


def test_order_field_value_forbidden_on_object():
    with pytest.raises(ForbiddenFieldArgumentError, match=r".*\"value\".*\"order\".*"):

        @strawberry_django.order_field
        def field_method(prefix, queryset, value: auto):
            pass

        @strawberry_django.order_field
        def order(prefix, queryset, value: auto):
            pass


def test_order_field_on_object():
    try:

        @strawberry_django.order_field
        def order(self, root, info: Info, prefix, sequence, queryset):
            pass

    except Exception as exc:
        raise pytest.fail(f"DID RAISE {exc}")


def test_order_field_method():
    @strawberry_django.ordering.order(models.Fruit)
    class Order:
        @strawberry_django.order_field
        def custom_order(
            self, root, info: Info, prefix, value: auto, sequence, queryset
        ):
            assert self == order, "Unexpected self passed"
            assert root == order, "Unexpected root passed"
            assert info == fake_info, "Unexpected info passed"
            assert prefix == "ROOT", "Unexpected prefix passed"
            assert value == Ordering.ASC, "Unexpected value passed"
            assert sequence == sequence_inner, "Unexpected sequence passed"
            assert queryset == qs, "Unexpected queryset passed"
            raise Exception("WAS CALLED")

    order = cast("WithStrawberryObjectDefinition", Order(custom_order=Ordering.ASC))  # type: ignore
    schema = strawberry.Schema(query=Query)
    fake_info: Any = type("FakeInfo", (), {"schema": schema})
    qs: Any = object()
    sequence_inner: Any = object()
    sequence = {"customOrder": OrderSequence(0, children=sequence_inner)}

    with pytest.raises(Exception, match="WAS CALLED"):
        process_order(order, fake_info, qs, prefix="ROOT", sequence=sequence)


def test_order_method_not_called_when_not_decorated(mocker: MockFixture):
    @strawberry_django.ordering.order(models.Fruit)
    class Order:
        def order(self, root, info: Info, prefix, value: auto, sequence, queryset):
            pytest.fail("Should not have been called")

    mock_order_method = mocker.spy(Order, "order")

    process_order(
        cast("WithStrawberryObjectDefinition", Order()), mock.Mock(), mock.Mock()
    )

    mock_order_method.assert_not_called()


def test_order_field_not_called(mocker: MockFixture):
    @strawberry_django.ordering.order(models.Fruit)
    class Order:
        order: Ordering = Ordering.ASC

    # Calling this and no error being raised is the test, as the wrong behavior would
    # be for the field to be called like a method
    process_order(
        cast("WithStrawberryObjectDefinition", Order()), mock.Mock(), mock.Mock()
    )


def test_order_object_method():
    @strawberry_django.ordering.order(models.Fruit)
    class Order:
        @strawberry_django.order_field
        def order(self, root, info: Info, prefix, sequence, queryset):
            assert self == order_, "Unexpected self passed"
            assert root == order_, "Unexpected root passed"
            assert info == fake_info, "Unexpected info passed"
            assert prefix == "ROOT", "Unexpected prefix passed"
            assert sequence == sequence_, "Unexpected sequence passed"
            assert queryset == qs, "Unexpected queryset passed"
            return queryset, ["name"]

    order_ = cast("WithStrawberryObjectDefinition", Order())
    schema = strawberry.Schema(query=Query)
    fake_info: Any = type("FakeInfo", (), {"schema": schema})
    qs: Any = object()
    sequence_: Any = {"customOrder": OrderSequence(0)}

    order = process_order(order_, fake_info, qs, prefix="ROOT", sequence=sequence_)[1]
    assert "name" in order, "order was not called"


def test_order_nulls(query, db, fruits):
    t1 = models.FruitType.objects.create(name="Type1")
    t2 = models.FruitType.objects.create(name="Type2")

    f1, f2, f3 = models.Fruit.objects.all()

    f2.types.add(t1)
    f3.types.add(t1, t2)

    result = query("{ fruits(order: { typesNumber: ASC }) { id } }")
    assert not result.errors
    assert result.data["fruits"] == [
        {"id": str(f1.id)},
        {"id": str(f2.id)},
        {"id": str(f3.id)},
    ]

    result = query("{ fruits(order: { typesNumber: DESC }) { id } }")
    assert not result.errors
    assert result.data["fruits"] == [
        {"id": str(f3.id)},
        {"id": str(f2.id)},
        {"id": str(f1.id)},
    ]

    result = query("{ fruits(order: { typesNumber: ASC_NULLS_FIRST }) { id } }")
    assert not result.errors
    assert result.data["fruits"] == [
        {"id": str(f1.id)},
        {"id": str(f2.id)},
        {"id": str(f3.id)},
    ]

    result = query("{ fruits(order: { typesNumber: ASC_NULLS_LAST }) { id } }")
    assert not result.errors
    assert result.data["fruits"] == [
        {"id": str(f2.id)},
        {"id": str(f3.id)},
        {"id": str(f1.id)},
    ]

    result = query("{ fruits(order: { typesNumber: DESC_NULLS_LAST }) { id } }")
    assert not result.errors
    assert result.data["fruits"] == [
        {"id": str(f3.id)},
        {"id": str(f2.id)},
        {"id": str(f1.id)},
    ]

    result = query("{ fruits(order: { typesNumber: DESC_NULLS_FIRST }) { id } }")
    assert not result.errors
    assert result.data["fruits"] == [
        {"id": str(f1.id)},
        {"id": str(f3.id)},
        {"id": str(f2.id)},
    ]



import pytest

from tests.utils import assert_num_queries

from .models import ArtProject, ResearchProject
from .schema import schema


@pytest.mark.django_db(transaction=True)
def test_polymorphic_interface_query():
    ap = ArtProject.objects.create(topic="Art", artist="Artist")
    rp = ResearchProject.objects.create(topic="Research", supervisor="Supervisor")

    query = """\
    query {
      projects {
        edges {
          node {
            __typename
            topic
            ... on ArtProjectType {
              artist
            }
            ... on ResearchProjectType {
              supervisor
            }
          }
        }
      }
    }
    """

    # ContentType, base table, two subtables = 4 queries
    with assert_num_queries(4):
        result = schema.execute_sync(query)
    assert not result.errors
    assert result.data == {
        "projects": {
            "edges": [
                {
                    "node": {
                        "__typename": "ArtProjectType",
                        "topic": ap.topic,
                        "artist": ap.artist,
                    }
                },
                {
                    "node": {
                        "__typename": "ResearchProjectType",
                        "topic": rp.topic,
                        "supervisor": rp.supervisor,
                    }
                },
            ]
        }
    }



from typing import Literal, TypeAlias

import pytest
from django.contrib.auth.models import Permission
from guardian.shortcuts import assign_perm
from strawberry.relay import to_base64

from strawberry_django.optimizer import DjangoOptimizerExtension

from .projects.faker import (
    GroupFactory,
    IssueFactory,
    MilestoneFactory,
    StaffUserFactory,
    SuperuserUserFactory,
    UserFactory,
)
from .utils import GraphQLTestClient, assert_num_queries

PermKind: TypeAlias = Literal["user", "group", "superuser"]
perm_kinds: list[PermKind] = ["user", "group", "superuser"]


@pytest.mark.django_db(transaction=True)
def test_is_authenticated(db, gql_client: GraphQLTestClient):
    query = """
    query Issue ($id: ID!) {
        issueLoginRequired (id: $id) {
          id
          name
        }
      }
    """
    issue = IssueFactory.create()

    res = gql_client.query(
        query,
        {"id": to_base64("IssueType", issue.pk)},
        assert_no_errors=False,
    )
    assert res.data is None
    assert res.errors == [
        {
            "message": "User is not authenticated.",
            "locations": [{"line": 3, "column": 9}],
            "path": ["issueLoginRequired"],
        },
    ]

    user = UserFactory.create()
    with gql_client.login(user):
        res = gql_client.query(query, {"id": to_base64("IssueType", issue.pk)})
        assert res.data == {
            "issueLoginRequired": {
                "id": to_base64("IssueType", issue.pk),
                "name": issue.name,
            },
        }


@pytest.mark.django_db(transaction=True)
def test_is_authenticated_optional(db, gql_client: GraphQLTestClient):
    query = """
    query Issue ($id: ID!) {
        issueLoginRequiredOptional (id: $id) {
          id
          name
        }
      }
    """
    issue = IssueFactory.create()

    res = gql_client.query(query, {"id": to_base64("IssueType", issue.pk)})
    assert res.data == {"issueLoginRequiredOptional": None}

    user = UserFactory.create()
    with gql_client.login(user):
        res = gql_client.query(query, {"id": to_base64("IssueType", issue.pk)})
        assert res.data == {
            "issueLoginRequiredOptional": {
                "id": to_base64("IssueType", issue.pk),
                "name": issue.name,
            },
        }


@pytest.mark.django_db(transaction=True)
def test_staff_required(db, gql_client: GraphQLTestClient):
    query = """
    query Issue ($id: ID!) {
        issueStaffRequired (id: $id) {
          id
          name
        }
      }
    """
    issue = IssueFactory.create()

    res = gql_client.query(
        query,
        {"id": to_base64("IssueType", issue.pk)},
        assert_no_errors=False,
    )
    assert res.data is None
    assert res.errors == [
        {
            "message": "User is not a staff member.",
            "locations": [{"line": 3, "column": 9}],
            "path": ["issueStaffRequired"],
        },
    ]

    user = UserFactory.create()
    with gql_client.login(user):
        assert res.data is None
        assert res.errors == [
            {
                "message": "User is not a staff member.",
                "locations": [{"line": 3, "column": 9}],
                "path": ["issueStaffRequired"],
            },
        ]

    staff = StaffUserFactory.create()
    with gql_client.login(staff):
        res = gql_client.query(query, {"id": to_base64("IssueType", issue.pk)})
        assert res.data == {
            "issueStaffRequired": {
                "id": to_base64("IssueType", issue.pk),
                "name": issue.name,
            },
        }


@pytest.mark.django_db(transaction=True)
def test_staff_required_optional(db, gql_client: GraphQLTestClient):
    query = """
    query Issue ($id: ID!) {
        issueStaffRequiredOptional (id: $id) {
          id
          name
        }
      }
    """
    issue = IssueFactory.create()

    res = gql_client.query(query, {"id": to_base64("IssueType", issue.pk)})
    assert res.data == {"issueStaffRequiredOptional": None}

    user = UserFactory.create()
    with gql_client.login(user):
        res = gql_client.query(query, {"id": to_base64("IssueType", issue.pk)})
        assert res.data == {"issueStaffRequiredOptional": None}

    staff = StaffUserFactory.create()
    with gql_client.login(staff):
        res = gql_client.query(query, {"id": to_base64("IssueType", issue.pk)})
        assert res.data == {
            "issueStaffRequiredOptional": {
                "id": to_base64("IssueType", issue.pk),
                "name": issue.name,
            },
        }


@pytest.mark.django_db(transaction=True)
def test_superuser_required(db, gql_client: GraphQLTestClient):
    query = """
    query Issue ($id: ID!) {
        issueSuperuserRequired (id: $id) {
          id
          name
        }
      }
    """
    issue = IssueFactory.create()

    res = gql_client.query(
        query,
        {"id": to_base64("IssueType", issue.pk)},
        assert_no_errors=False,
    )
    assert res.data is None
    assert res.errors == [
        {
            "message": "User is not a superuser.",
            "locations": [{"line": 3, "column": 9}],
            "path": ["issueSuperuserRequired"],
        },
    ]

    user = UserFactory.create()
    with gql_client.login(user):
        res = gql_client.query(
            query,
            {"id": to_base64("IssueType", issue.pk)},
            assert_no_errors=False,
        )
        assert res.data is None
        assert res.errors == [
            {
                "message": "User is not a superuser.",
                "locations": [{"line": 3, "column": 9}],
                "path": ["issueSuperuserRequired"],
            },
        ]

    superuser = SuperuserUserFactory.create()
    with gql_client.login(superuser):
        res = gql_client.query(query, {"id": to_base64("IssueType", issue.pk)})
        assert res.data == {
            "issueSuperuserRequired": {
                "id": to_base64("IssueType", issue.pk),
                "name": issue.name,
            },
        }


@pytest.mark.django_db(transaction=True)
def test_async_user_resolve(db, gql_client: GraphQLTestClient):
    query = """
    query asyncUserResolve {
        asyncUserResolve
      }
    """
    if not gql_client.is_async:
        pytest.skip("needs async client")
    user = UserFactory.create()
    with gql_client.login(user):
        res = gql_client.query(
            query,
            assert_no_errors=False,
        )
        assert res.data is None
        assert res.errors == [
            {
                "message": "You don't have permission to access this app.",
                "locations": [{"line": 3, "column": 9}],
                "path": ["asyncUserResolve"],
            },
        ]

    superuser = SuperuserUserFactory.create()
    with gql_client.login(superuser):
        res = gql_client.query(query)
        assert res.data == {"asyncUserResolve": True}
    user_with_perm = UserFactory.create()
    user_with_perm.user_permissions.add(
        Permission.objects.get(codename="view_issue"),
    )
    with gql_client.login(user_with_perm):
        res = gql_client.query(query)
        assert res.data == {"asyncUserResolve": True}


@pytest.mark.django_db(transaction=True)
def test_superuser_required_optional(db, gql_client: GraphQLTestClient):
    query = """
    query Issue ($id: ID!) {
        issueSuperuserRequiredOptional (id: $id) {
          id
          name
        }
      }
    """
    issue = IssueFactory.create()

    res = gql_client.query(query, {"id": to_base64("IssueType", issue.pk)})
    assert res.data == {"issueSuperuserRequiredOptional": None}

    user = UserFactory.create()
    with gql_client.login(user):
        res = gql_client.query(query, {"id": to_base64("IssueType", issue.pk)})
        assert res.data == {"issueSuperuserRequiredOptional": None}

    superuser = SuperuserUserFactory.create()
    with gql_client.login(superuser):
        res = gql_client.query(query, {"id": to_base64("IssueType", issue.pk)})
        assert res.data == {
            "issueSuperuserRequiredOptional": {
                "id": to_base64("IssueType", issue.pk),
                "name": issue.name,
            },
        }


@pytest.mark.django_db(transaction=True)
def test_perm_cached(db, gql_client: GraphQLTestClient):
    """Validates that the permission caching mechanism correctly stores permissions as a set of strings.

    The test targets the `_perm_cache` attribute used in `utils/query.py`. It verifies that
    the attribute behaves as expected, holding a `Set[str]` that represents permission
    codenames, rather than direct Permission objects.

    This test addresses a regression captured by the following error:

    ```
    user_perms: Set[str] = {p.codename for p in perm_cache}
                          ^^^^^^^^^^
    AttributeError: 'str' object has no attribute 'codename'
    ```
    """
    query = """
    query Issue ($id: ID!) {
        issuePermRequired (id: $id) {
          id
          privateName
        }
    }
    """
    issue = IssueFactory.create(name="Test")

    # User with permission
    user_with_perm = UserFactory.create()
    user_with_perm.user_permissions.add(
        Permission.objects.get(codename="view_issue"),
    )
    assign_perm("view_issue", user_with_perm, issue)
    with gql_client.login(user_with_perm):
        if DjangoOptimizerExtension.enabled.get():
            res = gql_client.query(query, {"id": to_base64("IssueType", issue.pk)})
            assert res.data == {
                "issuePermRequired": {
                    "id": to_base64("IssueType", issue.pk),
                    "privateName": issue.name,
                },
            }


@pytest.mark.django_db(transaction=True)
@pytest.mark.parametrize("kind", perm_kinds)
def test_perm_required(db, gql_client: GraphQLTestClient, kind: PermKind):
    query = """
    query Issue ($id: ID!) {
        issuePermRequired (id: $id) {
          id
          name
        }
      }
    """
    issue = IssueFactory.create()

    res = gql_client.query(
        query,
        {"id": to_base64("IssueType", issue.pk)},
        assert_no_errors=False,
    )
    assert res.data is None
    assert res.errors == [
        {
            "message": "You don't have permission to access this app.",
            "locations": [{"line": 3, "column": 9}],
            "path": ["issuePermRequired"],
        },
    ]

    user = UserFactory.create()
    with gql_client.login(user):
        res = gql_client.query(
            query,
            {"id": to_base64("IssueType", issue.pk)},
            assert_no_errors=False,
        )
        assert res.data is None
        assert res.errors == [
            {
                "message": "You don't have permission to access this app.",
                "locations": [{"line": 3, "column": 9}],
                "path": ["issuePermRequired"],
            },
        ]

    if kind == "user":
        user_with_perm = UserFactory.create()
        user_with_perm.user_permissions.add(
            Permission.objects.get(codename="view_issue"),
        )
    elif kind == "group":
        user_with_perm = UserFactory.create()
        group = GroupFactory.create()
        group.permissions.add(Permission.objects.get(codename="view_issue"))
        user_with_perm.groups.add(group)
    elif kind == "superuser":
        user_with_perm = SuperuserUserFactory.create()
    else:  # pragma:nocover
        raise AssertionError

    with gql_client.login(user_with_perm):
        res = gql_client.query(query, {"id": to_base64("IssueType", issue.pk)})
        assert res.data == {
            "issuePermRequired": {
                "id": to_base64("IssueType", issue.pk),
                "name": issue.name,
            },
        }


@pytest.mark.django_db(transaction=True)
@pytest.mark.parametrize("kind", perm_kinds)
def test_perm_required_optional(db, gql_client: GraphQLTestClient, kind: PermKind):
    query = """
    query Issue ($id: ID!) {
        issuePermRequiredOptional (id: $id) {
          id
          name
        }
      }
    """
    issue = IssueFactory.create()

    res = gql_client.query(query, {"id": to_base64("IssueType", issue.pk)})
    assert res.data == {"issuePermRequiredOptional": None}

    user = UserFactory.create()
    with gql_client.login(user):
        res = gql_client.query(query, {"id": to_base64("IssueType", issue.pk)})
        assert res.data == {"issuePermRequiredOptional": None}

    if kind == "user":
        user_with_perm = UserFactory.create()
        user_with_perm.user_permissions.add(
            Permission.objects.get(codename="view_issue"),
        )
    elif kind == "group":
        user_with_perm = UserFactory.create()
        group = GroupFactory.create()
        group.permissions.add(Permission.objects.get(codename="view_issue"))
        user_with_perm.groups.add(group)
    elif kind == "superuser":
        user_with_perm = SuperuserUserFactory.create()
    else:  # pragma:nocover
        raise AssertionError

    with gql_client.login(user_with_perm):
        res = gql_client.query(query, {"id": to_base64("IssueType", issue.pk)})
        assert res.data == {
            "issuePermRequiredOptional": {
                "id": to_base64("IssueType", issue.pk),
                "name": issue.name,
            },
        }


@pytest.mark.django_db(transaction=True)
@pytest.mark.parametrize("kind", perm_kinds)
def test_list_perm_required(db, gql_client: GraphQLTestClient, kind: PermKind):
    query = """
    query Issue {
        issueListPermRequired {
          id
          name
        }
      }
    """
    issue = IssueFactory.create()

    res = gql_client.query(query)
    assert res.data == {"issueListPermRequired": []}

    user = UserFactory.create()
    with gql_client.login(user):
        res = gql_client.query(query)
        assert res.data == {"issueListPermRequired": []}

    if kind == "user":
        user_with_perm = UserFactory.create()
        user_with_perm.user_permissions.add(
            Permission.objects.get(codename="view_issue"),
        )
    elif kind == "group":
        user_with_perm = UserFactory.create()
        group = GroupFactory.create()
        group.permissions.add(Permission.objects.get(codename="view_issue"))
        user_with_perm.groups.add(group)
    elif kind == "superuser":
        user_with_perm = SuperuserUserFactory.create()
    else:  # pragma:nocover
        raise AssertionError

    with gql_client.login(user_with_perm):
        res = gql_client.query(query)
        assert res.data == {
            "issueListPermRequired": [
                {
                    "id": to_base64("IssueType", issue.pk),
                    "name": issue.name,
                },
            ],
        }


@pytest.mark.django_db(transaction=True)
@pytest.mark.parametrize("kind", perm_kinds)
def test_conn_perm_required(db, gql_client: GraphQLTestClient, kind: PermKind):
    query = """
    query Issue {
        issueConnPermRequired {
          totalCount
          edges {
            node {
              id
              name
            }
          }
        }
      }
    """
    issue = IssueFactory.create()

    res = gql_client.query(query)
    assert res.data == {"issueConnPermRequired": {"edges": [], "totalCount": 0}}

    user = UserFactory.create()
    with gql_client.login(user):
        res = gql_client.query(query)
        assert res.data == {"issueConnPermRequired": {"edges": [], "totalCount": 0}}

    if kind == "user":
        user_with_perm = UserFactory.create()
        user_with_perm.user_permissions.add(
            Permission.objects.get(codename="view_issue"),
        )
    elif kind == "group":
        user_with_perm = UserFactory.create()
        group = GroupFactory.create()
        group.permissions.add(Permission.objects.get(codename="view_issue"))
        user_with_perm.groups.add(group)
    elif kind == "superuser":
        user_with_perm = SuperuserUserFactory.create()
    else:  # pragma:nocover
        raise AssertionError

    with gql_client.login(user_with_perm):
        res = gql_client.query(query)
        assert res.data == {
            "issueConnPermRequired": {
                "edges": [
                    {
                        "node": {
                            "id": to_base64("IssueType", issue.pk),
                            "name": issue.name,
                        },
                    },
                ],
                "totalCount": 1,
            },
        }


@pytest.mark.django_db(transaction=True)
@pytest.mark.parametrize("kind", perm_kinds)
def test_obj_perm_required(db, gql_client: GraphQLTestClient, kind: PermKind):
    query = """
    query Issue ($id: ID!) {
        issueObjPermRequired (id: $id) {
          id
          name
        }
      }
    """
    issue_no_perm = IssueFactory.create()
    issue_with_perm = IssueFactory.create()

    user = UserFactory.create()

    if kind == "user":
        user_with_perm = UserFactory.create()
        assign_perm("view_issue", user_with_perm, issue_with_perm)
    elif kind == "group":
        user_with_perm = UserFactory.create()
        group = GroupFactory.create()
        assign_perm("view_issue", group, issue_with_perm)
        user_with_perm.groups.add(group)
    elif kind == "superuser":
        user_with_perm = SuperuserUserFactory.create()
    else:  # pragma:nocover
        raise AssertionError

    for issue in [issue_no_perm, issue_with_perm]:
        res = gql_client.query(
            query,
            {"id": to_base64("IssueType", issue.pk)},
            assert_no_errors=False,
        )
        assert res.data is None
        assert res.errors == [
            {
                "message": "You don't have permission to access this app.",
                "locations": [{"line": 3, "column": 9}],
                "path": ["issueObjPermRequired"],
            },
        ]

    for u in [user, user_with_perm]:
        # Superusers will have access to everything...
        if kind == "superuser":
            continue

        with gql_client.login(u):
            res = gql_client.query(
                query,
                {"id": to_base64("IssueType", issue_no_perm.pk)},
                assert_no_errors=False,
            )
            assert res.data is None
            assert res.errors == [
                {
                    "message": "You don't have permission to access this app.",
                    "locations": [{"line": 3, "column": 9}],
                    "path": ["issueObjPermRequired"],
                },
            ]

    with gql_client.login(user_with_perm):
        res = gql_client.query(
            query,
            {"id": to_base64("IssueType", issue_with_perm.pk)},
        )
        assert res.data == {
            "issueObjPermRequired": {
                "id": to_base64("IssueType", issue_with_perm.pk),
                "name": issue_with_perm.name,
            },
        }


@pytest.mark.django_db(transaction=True)
@pytest.mark.parametrize("kind", perm_kinds)
def test_obj_perm_required_global(db, gql_client: GraphQLTestClient, kind: PermKind):
    query = """
    query Issue ($id: ID!) {
        issueObjPermRequired (id: $id) {
          id
          name
        }
      }
    """
    issue_no_perm = IssueFactory.create()
    issue_with_perm = IssueFactory.create()

    user = UserFactory.create()

    if kind == "user":
        user_with_perm = UserFactory.create()
        user_with_perm.user_permissions.add(
            Permission.objects.get(codename="view_issue"),
        )
    elif kind == "group":
        user_with_perm = UserFactory.create()
        group = GroupFactory.create()
        group.permissions.add(Permission.objects.get(codename="view_issue"))
        user_with_perm.groups.add(group)
    elif kind == "superuser":
        user_with_perm = SuperuserUserFactory.create()
    else:  # pragma:nocover
        raise AssertionError

    for issue in [issue_no_perm, issue_with_perm]:
        res = gql_client.query(
            query,
            {"id": to_base64("IssueType", issue.pk)},
            assert_no_errors=False,
        )
        assert res.data is None
        assert res.errors == [
            {
                "message": "You don't have permission to access this app.",
                "locations": [{"line": 3, "column": 9}],
                "path": ["issueObjPermRequired"],
            },
        ]

        with gql_client.login(user):
            res = gql_client.query(
                query,
                {"id": to_base64("IssueType", issue.pk)},
                assert_no_errors=False,
            )
            assert res.data is None
            assert res.errors == [
                {
                    "message": "You don't have permission to access this app.",
                    "locations": [{"line": 3, "column": 9}],
                    "path": ["issueObjPermRequired"],
                },
            ]

        with gql_client.login(user_with_perm):
            res = gql_client.query(
                query,
                {"id": to_base64("IssueType", issue_with_perm.pk)},
            )
            assert res.data == {
                "issueObjPermRequired": {
                    "id": to_base64("IssueType", issue_with_perm.pk),
                    "name": issue_with_perm.name,
                },
            }


@pytest.mark.django_db(transaction=True)
@pytest.mark.parametrize("kind", perm_kinds)
def test_obj_perm_required_optional(db, gql_client: GraphQLTestClient, kind: PermKind):
    query = """
    query Issue ($id: ID!) {
        issueObjPermRequiredOptional (id: $id) {
          id
          name
        }
      }
    """
    issue_no_perm = IssueFactory.create()
    issue_with_perm = IssueFactory.create()

    user = UserFactory.create()

    if kind == "user":
        user_with_perm = UserFactory.create()
        assign_perm("view_issue", user_with_perm, issue_with_perm)
    elif kind == "group":
        user_with_perm = UserFactory.create()
        group = GroupFactory.create()
        assign_perm("view_issue", group, issue_with_perm)
        user_with_perm.groups.add(group)
    elif kind == "superuser":
        user_with_perm = SuperuserUserFactory.create()
    else:  # pragma:nocover
        raise AssertionError

    for issue in [issue_no_perm, issue_with_perm]:
        res = gql_client.query(
            query,
            {"id": to_base64("IssueType", issue.pk)},
        )
        assert res.data == {"issueObjPermRequiredOptional": None}

    for u in [user, user_with_perm]:
        # Superusers will have access to everything...
        if kind == "superuser":
            continue

        with gql_client.login(u):
            res = gql_client.query(
                query,
                {"id": to_base64("IssueType", issue_no_perm.pk)},
            )
            assert res.data == {"issueObjPermRequiredOptional": None}

    with gql_client.login(user_with_perm):
        res = gql_client.query(
            query,
            {"id": to_base64("IssueType", issue_with_perm.pk)},
        )
        assert res.data == {
            "issueObjPermRequiredOptional": {
                "id": to_base64("IssueType", issue_with_perm.pk),
                "name": issue_with_perm.name,
            },
        }


@pytest.mark.django_db(transaction=True)
@pytest.mark.parametrize("kind", perm_kinds)
def test_list_obj_perm_required(db, gql_client: GraphQLTestClient, kind: PermKind):
    query = """
    query Issue {
        issueListObjPermRequired {
          id
          name
        }
      }
    """
    IssueFactory.create()
    issue_with_perm = IssueFactory.create()

    user = UserFactory.create()

    if kind == "user":
        user_with_perm = UserFactory.create()
        assign_perm("view_issue", user_with_perm, issue_with_perm)
    elif kind == "group":
        user_with_perm = UserFactory.create()
        group = GroupFactory.create()
        assign_perm("view_issue", group, issue_with_perm)
        user_with_perm.groups.add(group)
    elif kind == "superuser":
        user_with_perm = SuperuserUserFactory.create()
    else:  # pragma:nocover
        raise AssertionError

    res = gql_client.query(query)
    assert res.data == {"issueListObjPermRequired": []}

    with gql_client.login(user):
        res = gql_client.query(query)
        assert res.data == {"issueListObjPermRequired": []}

    if kind == "superuser":
        # Even though the user is a superuser, he doesn't have the permission
        # assigned directly to him for the listing.
        with gql_client.login(user_with_perm):
            res = gql_client.query(query)
            assert res.data == {"issueListObjPermRequired": []}
    else:
        with gql_client.login(user_with_perm):
            res = gql_client.query(query)
            assert res.data == {
                "issueListObjPermRequired": [
                    {
                        "id": to_base64("IssueType", issue_with_perm.pk),
                        "name": issue_with_perm.name,
                    },
                ],
            }


@pytest.mark.django_db(transaction=True)
@pytest.mark.parametrize("kind", perm_kinds)
def test_list_obj_perm_required_paginated(
    db, gql_client: GraphQLTestClient, kind: PermKind
):
    query = """
    query Issue {
        issueListObjPermRequiredPaginated(pagination: {limit: 10, offset: 0}) {
          id
          name
        }
      }
    """
    IssueFactory.create()
    issue_with_perm = IssueFactory.create()

    user = UserFactory.create()

    if kind == "user":
        user_with_perm = UserFactory.create()
        assign_perm("view_issue", user_with_perm, issue_with_perm)
    elif kind == "group":
        user_with_perm = UserFactory.create()
        group = GroupFactory.create()
        assign_perm("view_issue", group, issue_with_perm)
        user_with_perm.groups.add(group)
    elif kind == "superuser":
        user_with_perm = SuperuserUserFactory.create()
    else:  # pragma:nocover
        raise AssertionError

    res = gql_client.query(query)
    assert res.data == {"issueListObjPermRequiredPaginated": []}

    with gql_client.login(user):
        res = gql_client.query(query)
        assert res.data == {"issueListObjPermRequiredPaginated": []}

    if kind == "superuser":
        # Even though the user is a superuser, he doesn't have the permission
        # assigned directly to him for the listing.
        with gql_client.login(user_with_perm):
            res = gql_client.query(query)
            assert res.data == {"issueListObjPermRequiredPaginated": []}
    else:
        with gql_client.login(user_with_perm):
            res = gql_client.query(query)
            assert res.data == {
                "issueListObjPermRequiredPaginated": [
                    {
                        "id": to_base64("IssueType", issue_with_perm.pk),
                        "name": issue_with_perm.name,
                    },
                ],
            }


@pytest.mark.django_db(transaction=True)
@pytest.mark.parametrize("kind", perm_kinds)
def test_conn_obj_perm_required(db, gql_client: GraphQLTestClient, kind: PermKind):
    query = """
    query Issue {
        issueConnObjPermRequired {
          totalCount
          edges {
            node {
              id
              name
            }
          }
        }
      }
    """
    IssueFactory.create()
    issue_with_perm = IssueFactory.create()

    user = UserFactory.create()

    if kind == "user":
        user_with_perm = UserFactory.create()
        assign_perm("view_issue", user_with_perm, issue_with_perm)
    elif kind == "group":
        user_with_perm = UserFactory.create()
        group = GroupFactory.create()
        assign_perm("view_issue", group, issue_with_perm)
        user_with_perm.groups.add(group)
    elif kind == "superuser":
        user_with_perm = SuperuserUserFactory.create()
    else:  # pragma:nocover
        raise AssertionError

    res = gql_client.query(query)
    assert res.data == {"issueConnObjPermRequired": {"edges": [], "totalCount": 0}}

    with gql_client.login(user):
        res = gql_client.query(query)
        assert res.data == {"issueConnObjPermRequired": {"edges": [], "totalCount": 0}}

    if kind == "superuser":
        # Even though the user is a superuser, he doesn't have the permission
        # assigned directly to him for the listing.
        with gql_client.login(user_with_perm):
            res = gql_client.query(query)
            assert res.data == {
                "issueConnObjPermRequired": {"edges": [], "totalCount": 0},
            }
    else:
        with gql_client.login(user_with_perm):
            res = gql_client.query(query)
            assert res.data == {
                "issueConnObjPermRequired": {
                    "edges": [
                        {
                            "node": {
                                "id": to_base64("IssueType", issue_with_perm.pk),
                                "name": issue_with_perm.name,
                            },
                        },
                    ],
                    "totalCount": 1,
                },
            }


@pytest.mark.django_db(transaction=True)
def test_query_paginated_with_permissions(db, gql_client: GraphQLTestClient):
    query = """
      query TestQuery ($pagination: OffsetPaginationInput) {
        issuesPaginatedPermRequired (pagination: $pagination) {
          totalCount
          results {
            name
            milestone {
              name
            }
          }
        }
      }
    """

    milestone1 = MilestoneFactory.create()
    milestone2 = MilestoneFactory.create()

    issue1 = IssueFactory.create(milestone=milestone1)
    issue2 = IssueFactory.create(milestone=milestone1)
    issue3 = IssueFactory.create(milestone=milestone1)
    issue4 = IssueFactory.create(milestone=milestone2)
    issue5 = IssueFactory.create(milestone=milestone2)

    # No user logged in
    with assert_num_queries(0):
        res = gql_client.query(query)

    assert res.data == {
        "issuesPaginatedPermRequired": {
            "totalCount": 0,
            "results": [],
        }
    }

    user = UserFactory.create()

    # User logged in without permissions
    with gql_client.login(user):
        with assert_num_queries(4):
            res = gql_client.query(query)

        assert res.data == {
            "issuesPaginatedPermRequired": {
                "totalCount": 0,
                "results": [],
            }
        }

    # User logged in with permissions
    user.user_permissions.add(Permission.objects.get(codename="view_issue"))
    with gql_client.login(user):
        with assert_num_queries(6 if DjangoOptimizerExtension.enabled.get() else 11):
            res = gql_client.query(query)

        assert res.data == {
            "issuesPaginatedPermRequired": {
                "totalCount": 5,
                "results": [
                    {"name": issue1.name, "milestone": {"name": milestone1.name}},
                    {"name": issue2.name, "milestone": {"name": milestone1.name}},
                    {"name": issue3.name, "milestone": {"name": milestone1.name}},
                    {"name": issue4.name, "milestone": {"name": milestone2.name}},
                    {"name": issue5.name, "milestone": {"name": milestone2.name}},
                ],
            }
        }

        with assert_num_queries(6 if DjangoOptimizerExtension.enabled.get() else 8):
            res = gql_client.query(query, variables={"pagination": {"limit": 2}})

        assert res.data == {
            "issuesPaginatedPermRequired": {
                "totalCount": 5,
                "results": [
                    {"name": issue1.name, "milestone": {"name": milestone1.name}},
                    {"name": issue2.name, "milestone": {"name": milestone1.name}},
                ],
            }
        }


@pytest.mark.django_db(transaction=True)
def test_query_paginated_with_obj_permissions(db, gql_client: GraphQLTestClient):
    query = """
      query TestQuery ($pagination: OffsetPaginationInput) {
        issuesPaginatedObjPermRequired (pagination: $pagination) {
          totalCount
          results {
            name
            milestone {
              name
            }
          }
        }
      }
    """

    milestone1 = MilestoneFactory.create()
    milestone2 = MilestoneFactory.create()

    IssueFactory.create(milestone=milestone1)
    issue2 = IssueFactory.create(milestone=milestone1)
    IssueFactory.create(milestone=milestone1)
    issue4 = IssueFactory.create(milestone=milestone2)
    IssueFactory.create(milestone=milestone2)

    # No user logged in
    with assert_num_queries(0):
        res = gql_client.query(query)

    assert res.data == {
        "issuesPaginatedObjPermRequired": {
            "totalCount": 0,
            "results": [],
        }
    }

    user = UserFactory.create()

    # User logged in without permissions
    with gql_client.login(user):
        with assert_num_queries(5):
            res = gql_client.query(query)

        assert res.data == {
            "issuesPaginatedObjPermRequired": {
                "totalCount": 0,
                "results": [],
            }
        }

    assign_perm("view_issue", user, issue2)
    assign_perm("view_issue", user, issue4)

    # User logged in with permissions
    with gql_client.login(user):
        with assert_num_queries(4 if DjangoOptimizerExtension.enabled.get() else 6):
            res = gql_client.query(query)

        assert res.data == {
            "issuesPaginatedObjPermRequired": {
                "totalCount": 2,
                "results": [
                    {"name": issue2.name, "milestone": {"name": milestone1.name}},
                    {"name": issue4.name, "milestone": {"name": milestone2.name}},
                ],
            }
        }

        with assert_num_queries(4 if DjangoOptimizerExtension.enabled.get() else 5):
            res = gql_client.query(query, variables={"pagination": {"limit": 1}})

        assert res.data == {
            "issuesPaginatedObjPermRequired": {
                "totalCount": 2,
                "results": [
                    {"name": issue2.name, "milestone": {"name": milestone1.name}},
                ],
            }
        }



import pytest
import strawberry
from django.conf import settings
from django.contrib import auth as django_auth

from strawberry_django import auth
from tests import utils

from .test_mutations import User


@strawberry.type
class Query:
    current_user: User | None = auth.current_user()  # type: ignore


@pytest.fixture
def query(db):
    return utils.generate_query(Query)


def test_current_user(query, user, context):
    django_auth.login(
        context.request,
        user,
        backend=settings.AUTHENTICATION_BACKENDS[0],
    )

    result = query("{ currentUser { username } }", context_value=context)
    assert not result.errors
    assert result.data == {"currentUser": {"username": "user"}}


def test_current_user_not_logged_in(query, user, context):
    result = query("{ currentUser { username } }", context_value=context)
    assert result.errors
    assert result.data == {"currentUser": None}



import strawberry


@strawberry.type
class Query:
    @strawberry.field
    def hello(self, name: str | None = None) -> str:
        return f"Hello {name or 'world'}"


schema = strawberry.Schema(Query)



"""Custom GraphQL context and type definitions.

Provides type-safe context access and utilities for resolvers.
"""

from __future__ import annotations

import dataclasses
from typing import TYPE_CHECKING, Literal, TypeAlias, cast, overload

from asgiref.sync import sync_to_async
from django.core.exceptions import PermissionDenied
from django.utils.translation import gettext_lazy as _
from strawberry.django.context import StrawberryDjangoContext
from strawberry.types import info

if TYPE_CHECKING:
    from app.user.models import User

    from .dataloaders import DataLoaders


# Type alias for GraphQL info with our custom Context.
# Provides type safety and IDE autocompletion for context access.
Info: TypeAlias = info.Info["Context", None]


@dataclasses.dataclass
class Context(StrawberryDjangoContext):
    """Custom GraphQL context with type-safe user access and dataloaders.

    Extends StrawberryDjangoContext to add:
    - dataloaders: Container for all DataLoader instances
    - get_user(): Type-safe user retrieval with authentication checks
    - aget_user(): Async version of get_user()

    Example usage:
        @strawberry_django.field
        def my_resolver(self, info: Info) -> SomeType:
            user = info.context.get_user(required=True)
            brand = await info.context.dataloaders.brand_loader.load(brand_id)
    """

    dataloaders: DataLoaders

    @overload
    def get_user(self, *, required: Literal[True]) -> User: ...

    @overload
    def get_user(self, *, required: None = ...) -> User | None: ...

    def get_user(self, *, required: Literal[True] | None = None) -> User | None:
        """Get the authenticated user from the request.

        Args:
            required: If True, raises PermissionDenied when user is not authenticated.
                     If None/False, returns None for unauthenticated requests.

        Returns:
            User instance if authenticated, None otherwise (unless required=True)

        Raises:
            PermissionDenied: If required=True and user is not authenticated/active

        The overload signatures ensure type safety:
        - get_user(required=True) -> User (never None)
        - get_user() -> User | None

        """
        user = self.request.user

        if not user or not user.is_authenticated or not user.is_active:
            if required:
                raise PermissionDenied(_("No user logged in"))

            return None

        return cast("User", user)

    @overload
    async def aget_user(self, *, required: Literal[True]) -> User: ...

    @overload
    async def aget_user(self, *, required: None = ...) -> User | None: ...

    async def aget_user(self, *, required: Literal[True] | None = None) -> User | None:
        """Async version of get_user().

        Use this in async resolvers to avoid SynchronousOnlyOperation errors.

        Example:
            @strawberry_django.field
            async def my_async_resolver(self, info: Info) -> UserType:
                user = await info.context.aget_user(required=True)
                return cast("UserType", user)

        """
        # Wrap the sync method properly to handle the overload signature
        if required:
            return await sync_to_async(lambda: self.get_user(required=True))()
        return await sync_to_async(self.get_user)()



from __future__ import annotations

from strawberry import UNSET, auto, relay

import strawberry_django

from .models import Brand, Product, ProductImage


@strawberry_django.filter_type(Brand, lookups=True)
class BrandFilter:
    """Filter type for Brand queries with lookup support."""

    name: auto


@strawberry_django.type(
    Brand,
    name="Brand",
    filters=BrandFilter,
)
class BrandType(relay.Node):
    """GraphQL type for Brand model.

    Demonstrates reverse relationship (products) from ForeignKey.
    """

    name: auto
    products: list[ProductType]


@strawberry_django.filter_type(Product, lookups=True)
class ProductFilter:
    """Filter type for Product queries.

    Demonstrates nested filtering with the brand field, allowing queries like:
    filters: { brand: { name: { iContains: "apple" } } }
    """

    name: auto
    kind: auto
    brand: BrandFilter | None = UNSET


@strawberry_django.order_type(Product)
class ProductOrdering:
    """Ordering type for Product queries."""

    name: auto
    kind: auto


@strawberry_django.type(
    Product,
    name="Product",
    filters=ProductFilter,
    order=ProductOrdering,
)
class ProductType(relay.Node):
    """GraphQL type for Product model.

    Demonstrates:
    - Relay Node interface
    - Optional foreign key relationship (brand)
    - Enum field (kind)
    - Related objects list (images)
    - Custom computed fields
    """

    name: auto
    brand: BrandType | None
    kind: auto
    description: auto
    price: auto
    images: list[ProductImageType]

    @strawberry_django.field(only=["price"])
    def formatted_price(self, root: Product) -> str:
        """Return price formatted as currency string.

        Demonstrates a simple computed field for display formatting.
        """
        return f"${root.price:.2f}"


@strawberry_django.type(ProductImage, name="ProductImage")
class ProductImageType(relay.Node):
    """GraphQL type for ProductImage model."""

    @strawberry_django.field(only=["image"])
    def image(self, root: ProductImage) -> str | None:
        """Return the image URL if available."""
        return root.image.url if root.image else None


@strawberry_django.input(Product)
class ProductInput:
    """Input type for creating products."""

    name: auto
    brand: relay.GlobalID | None = None
    kind: auto
    description: auto
    price: auto


@strawberry_django.input(Brand)
class BrandInput:
    """Input type for creating brands."""

    name: auto



from typing import Any

import strawberry
from graphql import GraphQLAbstractType, GraphQLResolveInfo
from strawberry import Info
from strawberry.relay import Node

import strawberry_django
from strawberry_django.optimizer import DjangoOptimizerExtension
from strawberry_django.pagination import OffsetPaginated
from strawberry_django.relay import DjangoListConnection

from .models import Company, Project


@strawberry_django.interface(Project)
class ProjectType(Node):
    topic: strawberry.auto

    @classmethod
    def resolve_type(
        cls, value: Any, info: GraphQLResolveInfo, parent_type: GraphQLAbstractType
    ) -> str:
        if not isinstance(value, Project):
            raise TypeError
        if value.artist:
            return "ArtProjectType"
        if value.supervisor:
            return "ResearchProjectType"
        raise TypeError

    @classmethod
    def get_queryset(cls, qs, info: Info):
        return qs


@strawberry_django.type(Project)
class ArtProjectType(ProjectType):
    artist: strawberry.auto


@strawberry_django.type(Project)
class ResearchProjectType(ProjectType):
    supervisor: strawberry.auto


@strawberry_django.type(Company)
class CompanyType:
    name: strawberry.auto
    main_project: ProjectType | None
    projects: list[ProjectType]


@strawberry.type
class Query:
    companies: list[CompanyType] = strawberry_django.field()
    projects: list[ProjectType] = strawberry_django.field()
    projects_paginated: list[ProjectType] = strawberry_django.field(pagination=True)
    projects_offset_paginated: OffsetPaginated[ProjectType] = (
        strawberry_django.offset_paginated()
    )
    projects_connection: DjangoListConnection[ProjectType] = (
        strawberry_django.connection()
    )


schema = strawberry.Schema(
    query=Query,
    types=[ArtProjectType, ResearchProjectType],
    extensions=[DjangoOptimizerExtension],
)



from typing import Any, ClassVar, Generic, TypeVar

import factory
from django.contrib.auth import get_user_model
from django.contrib.auth.hashers import make_password
from django.contrib.auth.models import Group
from factory.declarations import Iterator, LazyFunction, Sequence, SubFactory
from factory.faker import Faker

from .models import Favorite, Issue, Milestone, Project, Quiz, Tag

_T = TypeVar("_T")
User = get_user_model()


class _BaseFactory(factory.django.DjangoModelFactory, Generic[_T]):
    Meta: ClassVar[Any]

    @classmethod
    def create(cls, **kwargs) -> _T:
        return super().create(**kwargs)

    @classmethod
    def create_batch(cls, size: int, **kwargs) -> list[_T]:
        return super().create_batch(size, **kwargs)


class GroupFactory(_BaseFactory[Group]):
    class Meta:
        model = Group

    name = Sequence(lambda n: f"Group {n}")


class UserFactory(_BaseFactory["User"]):
    class Meta:
        model = User

    is_active = True
    first_name = Faker("first_name")
    last_name = Faker("last_name")
    username = Sequence(lambda n: f"user-{n}")
    email = Faker("email")
    password = LazyFunction(lambda: make_password("foobar"))


class StaffUserFactory(UserFactory):
    is_staff = True


class SuperuserUserFactory(UserFactory):
    is_superuser = True


class ProjectFactory(_BaseFactory[Project]):
    class Meta:
        model = Project

    name = Sequence(lambda n: f"Project {n}")
    due_date = Faker("future_date")


class MilestoneFactory(_BaseFactory[Milestone]):
    class Meta:
        model = Milestone

    name = Sequence(lambda n: f"Milestone {n}")
    due_date = Faker("future_date")
    project = SubFactory(ProjectFactory)


class FavoriteFactory(_BaseFactory[Favorite]):
    class Meta:
        model = Favorite

    name = Sequence(lambda n: f"Favorite {n}")


class IssueFactory(_BaseFactory[Issue]):
    class Meta:
        model = Issue

    name = Sequence(lambda n: f"Issue {n}")
    kind = Iterator(Issue.Kind)
    milestone = SubFactory(MilestoneFactory)
    priority = Faker("pyint", min_value=0, max_value=5)


class TagFactory(_BaseFactory[Tag]):
    class Meta:
        model = Tag

    name = Sequence(lambda n: f"Tag {n}")


class QuizFactory(_BaseFactory[Quiz]):
    class Meta:
        model = Quiz

    title = Sequence(lambda n: f"Quiz {n}")



from typing import cast

import pytest
import strawberry
from django.db.models import CharField, Q, QuerySet, Value
from django.db.models.functions import Concat
from strawberry import auto
from strawberry.types.base import WithStrawberryObjectDefinition

import strawberry_django
from strawberry_django.fields import filter_types
from strawberry_django.filters import process_filters
from tests import models


def test_process_filters_prefix(db):
    @strawberry_django.filter_type(models.Vegetable)
    class VegetableFilter:
        name: auto

        @strawberry_django.filter_field
        def name_and_description(
            self,
            info: strawberry.Info,
            queryset: QuerySet,
            value: filter_types.StrFilterLookup,
            prefix: str,
        ) -> tuple[QuerySet, Q]:
            queryset = queryset.alias(
                _name_description=Concat(
                    f"{prefix}name",
                    Value(" "),
                    f"{prefix}description",
                    output_field=CharField(),
                )
            )
            return process_filters(
                cast("WithStrawberryObjectDefinition", value),
                queryset,
                info,
                prefix="_name_description__",
            )

    @strawberry_django.type(models.Vegetable, filters=VegetableFilter)
    class VegetableType:
        id: auto
        name: auto

    @strawberry.type
    class Query:
        vegetables: list[VegetableType] = strawberry_django.field()

    schema = strawberry.Schema(query=Query)

    models.Vegetable.objects.create(
        name="carrot", description="orange root", world_production=40.0e6
    )
    result = schema.execute_sync("""
    {
        vegetables(filters: {
            nameAndDescription: { iContains: "orange" }
        }) { name }
    }
    """)
    assert not result.errors
    assert result.data == {"vegetables": [{"name": "carrot"}]}


def test_process_filters_prefix_missing_trailing_underscores(db):
    @strawberry_django.filter_type(models.Vegetable)
    class VegetableFilter:
        name: auto

        @strawberry_django.filter_field
        def name_and_description(
            self,
            info: strawberry.Info,
            queryset: QuerySet,
            value: filter_types.StrFilterLookup,
            prefix: str,
        ) -> tuple[QuerySet, Q]:
            queryset = queryset.alias(
                _name_description=Concat(
                    f"{prefix}name",
                    Value(" "),
                    f"{prefix}description",
                    output_field=CharField(),
                )
            )
            return process_filters(
                cast("WithStrawberryObjectDefinition", value),
                queryset,
                info,
                prefix="_name_description",
            )

    @strawberry_django.type(models.Vegetable, filters=VegetableFilter)
    class VegetableType:
        id: auto
        name: auto

    @strawberry.type
    class Query:
        vegetables: list[VegetableType] = strawberry_django.field()

    schema = strawberry.Schema(query=Query)

    models.Vegetable.objects.create(
        name="carrot", description="orange root", world_production=40.0e6
    )
    with pytest.warns(UserWarning, match="does not end with '__'"):
        result = schema.execute_sync("""
        {
            vegetables(filters: {
                nameAndDescription: { iContains: "orange" }
            }) { name }
        }
        """)
    assert result.errors



"""Tests for strawberry_django.federation.type decorator."""

from typing import Any, cast

import pytest
import strawberry
from strawberry.federation.schema_directives import (
    Authenticated,
    Inaccessible,
    Key,
    Policy,
    RequiresScopes,
    Shareable,
    Tag,
)
from strawberry.federation.types import FieldSet
from strawberry.types import get_object_definition

import strawberry_django
from tests import models

# =============================================================================
# Tests for federation type decorator with keys
# =============================================================================


def test_federation_type_with_string_key():
    """Test that @strawberry_django.federation.type accepts string keys."""

    @strawberry_django.federation.type(models.Fruit, keys=["id"])
    class FruitType:
        id: strawberry.auto
        name: strawberry.auto

    type_def = get_object_definition(FruitType, strict=True)
    key_directives = [d for d in type_def.directives or [] if isinstance(d, Key)]
    assert len(key_directives) == 1
    assert str(key_directives[0].fields) == "id"


def test_federation_type_with_key_directive_object():
    """Test using Key directive directly for complex keys."""

    @strawberry_django.federation.type(
        models.Fruit,
        keys=[Key(fields=FieldSet("id"), resolvable=False)],
    )
    class FruitType:
        id: strawberry.auto

    type_def = get_object_definition(FruitType, strict=True)
    key_directive = next(d for d in type_def.directives or [] if isinstance(d, Key))
    assert key_directive.resolvable is False


@pytest.mark.parametrize(
    ("keys", "expected_fields"),
    [
        (["id"], {"id"}),
        (["id", "name"], {"id", "name"}),
        (["id", "name", "sweetness"], {"id", "name", "sweetness"}),
    ],
)
def test_federation_type_with_multiple_keys(keys, expected_fields):
    """Test type with multiple @key directives."""

    @strawberry_django.federation.type(models.Fruit, keys=keys)
    class FruitType:
        id: strawberry.auto
        name: strawberry.auto
        sweetness: strawberry.auto

    type_def = get_object_definition(FruitType, strict=True)
    key_directives = [d for d in type_def.directives or [] if isinstance(d, Key)]
    assert len(key_directives) == len(keys)
    key_fields = {str(k.fields) for k in key_directives}
    assert key_fields == expected_fields


def test_federation_type_with_composite_key():
    """Test type with composite @key directive (multiple fields in one key)."""

    @strawberry_django.federation.type(models.Fruit, keys=["id name"])
    class FruitType:
        id: strawberry.auto
        name: strawberry.auto

    type_def = get_object_definition(FruitType, strict=True)
    key_directives = [d for d in type_def.directives or [] if isinstance(d, Key)]
    assert len(key_directives) == 1
    assert str(key_directives[0].fields) == "id name"


# =============================================================================
# Tests for federation directives
# =============================================================================


@pytest.mark.parametrize(
    ("directive_param", "directive_value", "directive_class"),
    [
        ("shareable", True, Shareable),
        ("authenticated", True, Authenticated),
        ("inaccessible", True, Inaccessible),
    ],
)
def test_federation_type_boolean_directives(
    directive_param, directive_value, directive_class
):
    """Test boolean federation directives (@shareable, @authenticated, @inaccessible)."""
    kwargs = cast("dict[str, Any]", {"keys": ["id"], directive_param: directive_value})

    @strawberry_django.federation.type(models.Fruit, **kwargs)
    class FruitType:
        id: strawberry.auto

    type_def = get_object_definition(FruitType, strict=True)
    assert any(isinstance(d, directive_class) for d in type_def.directives or [])


def test_federation_type_inaccessible_false_no_directive():
    """Test that inaccessible=False does NOT add @inaccessible directive."""

    @strawberry_django.federation.type(models.Fruit, keys=["id"], inaccessible=False)
    class FruitType:
        id: strawberry.auto

    type_def = get_object_definition(FruitType, strict=True)
    assert not any(isinstance(d, Inaccessible) for d in type_def.directives or [])


def test_federation_type_with_policy():
    """Test @policy directive."""

    @strawberry_django.federation.type(
        models.Fruit,
        keys=["id"],
        policy=[["read:fruits", "admin"]],
    )
    class FruitType:
        id: strawberry.auto

    type_def = get_object_definition(FruitType, strict=True)
    policy_directives = [d for d in type_def.directives or [] if isinstance(d, Policy)]
    assert len(policy_directives) == 1
    assert policy_directives[0].policies == [["read:fruits", "admin"]]


def test_federation_type_with_requires_scopes():
    """Test @requiresScopes directive."""

    @strawberry_django.federation.type(
        models.Fruit,
        keys=["id"],
        requires_scopes=[["read:fruits"]],
    )
    class FruitType:
        id: strawberry.auto

    type_def = get_object_definition(FruitType, strict=True)
    scopes_directives = [
        d for d in type_def.directives or [] if isinstance(d, RequiresScopes)
    ]
    assert len(scopes_directives) == 1
    assert scopes_directives[0].scopes == [["read:fruits"]]


@pytest.mark.parametrize(
    "tags",
    [
        ["internal"],
        ["internal", "deprecated"],
        ["v1", "public", "stable"],
    ],
)
def test_federation_type_with_tags(tags):
    """Test @tag directives with various tag counts."""

    @strawberry_django.federation.type(models.Fruit, keys=["id"], tags=tags)
    class FruitType:
        id: strawberry.auto

    type_def = get_object_definition(FruitType, strict=True)
    tag_directives = [d for d in type_def.directives or [] if isinstance(d, Tag)]
    assert len(tag_directives) == len(tags)
    tag_names = {t.name for t in tag_directives}
    assert tag_names == set(tags)


def test_federation_type_with_extend():
    """Test that extend=True is passed through to the type definition."""

    @strawberry_django.federation.type(models.Fruit, keys=["id"], extend=True)
    class FruitType:
        id: strawberry.auto
        name: strawberry.auto

    type_def = get_object_definition(FruitType, strict=True)
    assert type_def.extend is True


def test_federation_type_without_keys():
    """Test that federation.type works without keys (non-entity with directives)."""

    @strawberry_django.federation.type(models.Fruit, shareable=True)
    class FruitType:
        id: strawberry.auto
        name: strawberry.auto

    type_def = get_object_definition(FruitType, strict=True)
    # Should have @shareable but no @key
    assert any(isinstance(d, Shareable) for d in type_def.directives or [])
    assert not any(isinstance(d, Key) for d in type_def.directives or [])


def test_federation_type_preserves_custom_directives():
    """Test that custom directives are preserved alongside federation directives."""
    from strawberry.schema_directive import Location

    @strawberry.schema_directive(locations=[Location.OBJECT])
    class CustomDirective:
        name: str

    @strawberry_django.federation.type(
        models.Fruit,
        keys=["id"],
        directives=[CustomDirective(name="test")],
    )
    class FruitType:
        id: strawberry.auto

    type_def = get_object_definition(FruitType, strict=True)
    # Both Key and custom directive should be present
    assert any(isinstance(d, Key) for d in type_def.directives or [])
    assert any(isinstance(d, CustomDirective) for d in type_def.directives or [])


# =============================================================================
# Tests for auto-generated resolve_reference
# =============================================================================


def test_federation_type_auto_generates_resolve_reference():
    """Test that resolve_reference is auto-generated for keyed types."""

    @strawberry_django.federation.type(models.Fruit, keys=["id"])
    class FruitType:
        id: strawberry.auto
        name: strawberry.auto

    assert hasattr(FruitType, "resolve_reference")
    assert callable(cast("Any", FruitType).resolve_reference)


def test_federation_type_does_not_override_custom_resolve_reference():
    """Test that custom resolve_reference is not overwritten."""

    @strawberry_django.federation.type(models.Fruit, keys=["id"])
    class FruitType:
        id: strawberry.auto
        name: strawberry.auto

        @classmethod
        def resolve_reference(cls, id: int):  # noqa: A002
            """Return the Fruit with given id."""
            return models.Fruit.objects.filter(id=id).first()

    # Verify custom method is preserved by checking docstring
    assert "Return the Fruit" in (
        cast("Any", FruitType).resolve_reference.__doc__ or ""
    )


def test_federation_type_without_keys_no_resolve_reference():
    """Test that types without keys don't get resolve_reference."""

    @strawberry_django.federation.type(models.Fruit, shareable=True)
    class FruitType:
        id: strawberry.auto
        name: strawberry.auto

    # Should NOT have auto-generated resolve_reference
    type_def = get_object_definition(FruitType, strict=True)
    assert not any(isinstance(d, Key) for d in type_def.directives or [])
    assert "resolve_reference" not in FruitType.__dict__


# =============================================================================
# Tests for Django feature integration
# =============================================================================


def test_federation_type_with_filters():
    """Test that filters work with federation types."""

    @strawberry_django.filters.filter(models.Fruit)
    class FruitFilter:
        name: strawberry.auto

    @strawberry_django.federation.type(
        models.Fruit,
        keys=["id"],
        filters=FruitFilter,
    )
    class FruitType:
        id: strawberry.auto
        name: strawberry.auto

    django_def = cast("Any", FruitType).__strawberry_django_definition__
    assert django_def.filters is FruitFilter


def test_federation_type_with_pagination():
    """Test that pagination works with federation types."""

    @strawberry_django.federation.type(
        models.Fruit,
        keys=["id"],
        pagination=True,
    )
    class FruitType:
        id: strawberry.auto
        name: strawberry.auto

    django_def = cast("Any", FruitType).__strawberry_django_definition__
    assert django_def.pagination is True


def test_federation_type_with_ordering():
    """Test that ordering works with federation types."""

    @strawberry_django.order_type(models.Fruit)
    class FruitOrder:
        name: strawberry.auto

    @strawberry_django.federation.type(
        models.Fruit,
        keys=["id"],
        ordering=FruitOrder,
    )
    class FruitType:
        id: strawberry.auto
        name: strawberry.auto

    django_def = cast("Any", FruitType).__strawberry_django_definition__
    assert django_def.ordering is FruitOrder


def test_federation_type_with_optimization_hints():
    """Test that optimization hints work with federation types."""

    @strawberry_django.federation.type(
        models.Fruit,
        keys=["id"],
        select_related=["color"],
        prefetch_related=["types"],
    )
    class FruitType:
        id: strawberry.auto
        name: strawberry.auto

    django_def = cast("Any", FruitType).__strawberry_django_definition__
    assert "color" in (django_def.store.select_related or [])
    assert any("types" in str(p) for p in (django_def.store.prefetch_related or []))


# =============================================================================
# Tests for federation interface decorator
# =============================================================================


def test_federation_interface_with_keys():
    """Test federation interface with @key directive."""

    @strawberry_django.federation.interface(models.Fruit, keys=["id"])
    class FruitInterface:
        id: strawberry.auto

    type_def = get_object_definition(FruitInterface, strict=True)
    assert type_def.is_interface
    assert any(isinstance(d, Key) for d in type_def.directives or [])


def test_federation_interface_auto_generates_resolve_reference():
    """Test that resolve_reference is auto-generated for keyed interfaces."""

    @strawberry_django.federation.interface(models.Fruit, keys=["id"])
    class FruitInterface:
        id: strawberry.auto

    assert hasattr(FruitInterface, "resolve_reference")
    assert callable(cast("Any", FruitInterface).resolve_reference)


@pytest.mark.parametrize(
    ("directive_param", "directive_value", "directive_class"),
    [
        ("authenticated", True, Authenticated),
        ("inaccessible", True, Inaccessible),
    ],
)
def test_federation_interface_directives(
    directive_param, directive_value, directive_class
):
    """Test federation directives on interfaces."""
    kwargs = cast("dict[str, Any]", {"keys": ["id"], directive_param: directive_value})

    @strawberry_django.federation.interface(models.Fruit, **kwargs)
    class FruitInterface:
        id: strawberry.auto

    type_def = get_object_definition(FruitInterface, strict=True)
    assert any(isinstance(d, directive_class) for d in type_def.directives or [])



from __future__ import annotations

import dataclasses

from app.product.models import Brand
from app.user.models import User
from asgiref.sync import sync_to_async
from strawberry.dataloader import DataLoader


async def load_brands(keys: list[int]) -> list[Brand | None]:
    """Batch load brands by their IDs."""
    brands = await sync_to_async(list)(Brand.objects.filter(id__in=keys))

    # Return results in the same order as keys
    brand_map = {brand.id: brand for brand in brands}
    return [brand_map.get(key) for key in keys]


async def load_users(keys: list[int]) -> list[User | None]:
    """Batch load users by their IDs."""
    users = await sync_to_async(list)(User.objects.filter(id__in=keys))

    # Return results in the same order as keys
    user_map = {user.id: user for user in users}
    return [user_map.get(key) for key in keys]


@dataclasses.dataclass
class DataLoaders:
    """Container for all dataloaders in the application.

    DataLoaders help solve the N+1 query problem by batching and caching database queries.
    Each loader is instantiated once per request and shared across all resolvers.
    """

    brand_loader: DataLoader[int, Brand | None] = dataclasses.field(
        default_factory=lambda: DataLoader(load_fn=load_brands)
    )
    user_loader: DataLoader[int, User | None] = dataclasses.field(
        default_factory=lambda: DataLoader(load_fn=load_users)
    )



from django.db import models


class RelayAuthor(models.Model):
    name = models.CharField(max_length=100)


class RelayBook(models.Model):
    title = models.CharField(max_length=100)
    author = models.ForeignKey(
        RelayAuthor,
        on_delete=models.CASCADE,
        related_name="books",
    )



from collections.abc import Awaitable
from typing import Any

import pytest
import strawberry
from strawberry import BasePermission, Info

import strawberry_django
from strawberry_django.optimizer import DjangoOptimizerExtension
from tests import models


@pytest.mark.django_db(transaction=True)
async def test_with_async_permission(db):
    class AsyncPermission(BasePermission):
        async def has_permission(  # type: ignore
            self,
            source: Any,
            info: Info,
            **kwargs: Any,
        ) -> bool | Awaitable[bool]:
            return True

    @strawberry_django.type(models.Fruit)
    class Fruit:
        name: strawberry.auto

    @strawberry_django.type(models.Color)
    class Color:
        name: strawberry.auto
        fruits: list[Fruit] = strawberry_django.field(
            permission_classes=[AsyncPermission]
        )

    @strawberry.type(name="Query")
    class Query:
        colors: list[Color] = strawberry_django.field()

    red = await models.Color.objects.acreate(name="Red")
    yellow = await models.Color.objects.acreate(name="Yellow")

    await models.Fruit.objects.acreate(name="Apple", color=red)
    await models.Fruit.objects.acreate(name="Banana", color=yellow)
    await models.Fruit.objects.acreate(name="Strawberry", color=red)

    schema = strawberry.Schema(query=Query)
    query = """
        query {
          colors {
            name
            fruits {
              name
            }
          }
        }
    """

    result = await schema.execute(query)
    assert result.errors is None
    assert result.data == {
        "colors": [
            {
                "name": "Red",
                "fruits": [
                    {"name": "Apple"},
                    {"name": "Strawberry"},
                ],
            },
            {
                "name": "Yellow",
                "fruits": [{"name": "Banana"}],
            },
        ]
    }


@pytest.mark.django_db(transaction=True)
async def test_with_async_permission_and_optimizer(db):
    class AsyncPermission(BasePermission):
        async def has_permission(  # type: ignore
            self,
            source: Any,
            info: Info,
            **kwargs: Any,
        ) -> bool | Awaitable[bool]:
            return True

    @strawberry_django.type(models.Fruit)
    class Fruit:
        name: strawberry.auto

    @strawberry_django.type(models.Color)
    class Color:
        name: strawberry.auto
        fruits: list[Fruit] = strawberry_django.field(
            permission_classes=[AsyncPermission]
        )

    @strawberry.type(name="Query")
    class Query:
        colors: list[Color] = strawberry_django.field()

    red = await models.Color.objects.acreate(name="Red")
    yellow = await models.Color.objects.acreate(name="Yellow")

    await models.Fruit.objects.acreate(name="Apple", color=red)
    await models.Fruit.objects.acreate(name="Banana", color=yellow)
    await models.Fruit.objects.acreate(name="Strawberry", color=red)

    schema = strawberry.Schema(
        query=Query,
        extensions=[DjangoOptimizerExtension()],
    )
    query = """
        query {
          colors {
            name
            fruits {
              name
            }
          }
        }
    """

    result = await schema.execute(query)
    assert result.errors is None
    assert result.data == {
        "colors": [
            {
                "name": "Red",
                "fruits": [
                    {"name": "Apple"},
                    {"name": "Strawberry"},
                ],
            },
            {
                "name": "Yellow",
                "fruits": [{"name": "Banana"}],
            },
        ]
    }


@pytest.mark.django_db(transaction=True)
def test_with_sync_permission(db):
    class AsyncPermission(BasePermission):
        def has_permission(
            self,
            source: Any,
            info: Info,
            **kwargs: Any,
        ) -> bool | Awaitable[bool]:
            return True

    @strawberry_django.type(models.Fruit)
    class Fruit:
        name: strawberry.auto

    @strawberry_django.type(models.Color)
    class Color:
        name: strawberry.auto
        fruits: list[Fruit] = strawberry_django.field(
            permission_classes=[AsyncPermission]
        )

    @strawberry.type(name="Query")
    class Query:
        colors: list[Color] = strawberry_django.field()

    red = models.Color.objects.create(name="Red")
    yellow = models.Color.objects.create(name="Yellow")

    models.Fruit.objects.create(name="Apple", color=red)
    models.Fruit.objects.create(name="Banana", color=yellow)
    models.Fruit.objects.create(name="Strawberry", color=red)

    schema = strawberry.Schema(query=Query)
    query = """
        query {
          colors {
            name
            fruits {
              name
            }
          }
        }
    """

    result = schema.execute_sync(query)
    assert result.errors is None
    assert result.data == {
        "colors": [
            {
                "name": "Red",
                "fruits": [
                    {"name": "Apple"},
                    {"name": "Strawberry"},
                ],
            },
            {
                "name": "Yellow",
                "fruits": [{"name": "Banana"}],
            },
        ]
    }


@pytest.mark.django_db(transaction=True)
def test_with_sync_permission_and_optimizer(db):
    class AsyncPermission(BasePermission):
        def has_permission(
            self,
            source: Any,
            info: Info,
            **kwargs: Any,
        ) -> bool | Awaitable[bool]:
            return True

    @strawberry_django.type(models.Fruit)
    class Fruit:
        name: strawberry.auto

    @strawberry_django.type(models.Color)
    class Color:
        name: strawberry.auto
        fruits: list[Fruit] = strawberry_django.field(
            permission_classes=[AsyncPermission]
        )

    @strawberry.type(name="Query")
    class Query:
        colors: list[Color] = strawberry_django.field()

    red = models.Color.objects.create(name="Red")
    yellow = models.Color.objects.create(name="Yellow")

    models.Fruit.objects.create(name="Apple", color=red)
    models.Fruit.objects.create(name="Banana", color=yellow)
    models.Fruit.objects.create(name="Strawberry", color=red)

    schema = strawberry.Schema(
        query=Query,
        extensions=[DjangoOptimizerExtension()],
    )
    query = """
        query {
          colors {
            name
            fruits {
              name
            }
          }
        }
    """

    result = schema.execute_sync(query)
    assert result.errors is None
    assert result.data == {
        "colors": [
            {
                "name": "Red",
                "fruits": [
                    {"name": "Apple"},
                    {"name": "Strawberry"},
                ],
            },
            {
                "name": "Yellow",
                "fruits": [{"name": "Banana"}],
            },
        ]
    }



import textwrap
from typing import Annotated

import pytest
import strawberry
from django.db.models import QuerySet
from django.test.utils import override_settings
from strawberry.extensions.field_extension import FieldExtension

import strawberry_django
from strawberry_django.optimizer import DjangoOptimizerExtension
from strawberry_django.pagination import OffsetPaginated, OffsetPaginationInput
from strawberry_django.settings import StrawberryDjangoSettings
from tests import models


def test_paginated_schema():
    @strawberry_django.type(models.Fruit)
    class Fruit:
        id: int
        name: str

    @strawberry_django.type(models.Color)
    class Color:
        id: int
        name: str
        fruits: OffsetPaginated[Fruit]

    @strawberry.type
    class Query:
        fruits: OffsetPaginated[Fruit] = strawberry_django.offset_paginated()
        colors: OffsetPaginated[Color] = strawberry_django.offset_paginated()

    schema = strawberry.Schema(query=Query)

    expected = '''\
    type Color {
      id: Int!
      name: String!
      fruits(pagination: OffsetPaginationInput): FruitOffsetPaginated!
    }

    type ColorOffsetPaginated {
      pageInfo: OffsetPaginationInfo!

      """Total count of existing results."""
      totalCount: Int!

      """List of paginated results."""
      results: [Color!]!
    }

    type Fruit {
      id: Int!
      name: String!
    }

    type FruitOffsetPaginated {
      pageInfo: OffsetPaginationInfo!

      """Total count of existing results."""
      totalCount: Int!

      """List of paginated results."""
      results: [Fruit!]!
    }

    type OffsetPaginationInfo {
      offset: Int!
      limit: Int
    }

    input OffsetPaginationInput {
      offset: Int! = 0
      limit: Int
    }

    type Query {
      fruits(pagination: OffsetPaginationInput): FruitOffsetPaginated!
      colors(pagination: OffsetPaginationInput): ColorOffsetPaginated!
    }
    '''

    assert textwrap.dedent(str(schema)) == textwrap.dedent(expected).strip()


@pytest.mark.django_db(transaction=True)
def test_pagination_query():
    @strawberry_django.type(models.Fruit)
    class Fruit:
        id: int
        name: str

    @strawberry.type
    class Query:
        fruits: OffsetPaginated[Fruit] = strawberry_django.offset_paginated()

    models.Fruit.objects.create(name="Apple")
    models.Fruit.objects.create(name="Banana")
    models.Fruit.objects.create(name="Strawberry")

    schema = strawberry.Schema(query=Query)

    query = """\
    query GetFruits ($pagination: OffsetPaginationInput) {
      fruits (pagination: $pagination) {
        totalCount
        results {
          name
        }
      }
    }
    """

    res = schema.execute_sync(query)
    assert res.errors is None
    assert res.data == {
        "fruits": {
            "totalCount": 3,
            "results": [{"name": "Apple"}, {"name": "Banana"}, {"name": "Strawberry"}],
        }
    }

    res = schema.execute_sync(query, variable_values={"pagination": {"limit": 1}})
    assert res.errors is None
    assert res.data == {
        "fruits": {
            "totalCount": 3,
            "results": [{"name": "Apple"}],
        }
    }

    res = schema.execute_sync(
        query, variable_values={"pagination": {"limit": 1, "offset": 1}}
    )
    assert res.errors is None
    assert res.data == {
        "fruits": {
            "totalCount": 3,
            "results": [{"name": "Banana"}],
        }
    }


@pytest.mark.django_db(transaction=True)
async def test_pagination_query_async():
    @strawberry_django.type(models.Fruit)
    class Fruit:
        id: int
        name: str

    @strawberry.type
    class Query:
        fruits: OffsetPaginated[Fruit] = strawberry_django.offset_paginated()

    await models.Fruit.objects.acreate(name="Apple")
    await models.Fruit.objects.acreate(name="Banana")
    await models.Fruit.objects.acreate(name="Strawberry")

    schema = strawberry.Schema(query=Query)

    query = """\
    query GetFruits ($pagination: OffsetPaginationInput) {
      fruits (pagination: $pagination) {
        totalCount
        results {
          name
        }
      }
    }
    """

    res = await schema.execute(query)
    assert res.errors is None
    assert res.data == {
        "fruits": {
            "totalCount": 3,
            "results": [{"name": "Apple"}, {"name": "Banana"}, {"name": "Strawberry"}],
        }
    }

    res = await schema.execute(query, variable_values={"pagination": {"limit": 1}})
    assert res.errors is None
    assert res.data == {
        "fruits": {
            "totalCount": 3,
            "results": [{"name": "Apple"}],
        }
    }

    res = await schema.execute(
        query, variable_values={"pagination": {"limit": 1, "offset": 1}}
    )
    assert res.errors is None
    assert res.data == {
        "fruits": {
            "totalCount": 3,
            "results": [{"name": "Banana"}],
        }
    }


@strawberry_django.type(models.Fruit)
class FruitLazyTest:
    id: int
    name: str


@strawberry_django.type(models.Color)
class ColorLazyTest:
    id: int
    name: str
    fruits: OffsetPaginated[
        Annotated["FruitLazyTest", strawberry.lazy("tests.test_paginated_type")]
    ] = strawberry_django.offset_paginated()


@pytest.mark.django_db(transaction=True)
def test_pagination_with_lazy_type_and_django_query_optimizer():
    @strawberry.type
    class Query:
        colors: OffsetPaginated[ColorLazyTest] = strawberry_django.offset_paginated()

    red = models.Color.objects.create(name="Red")
    yellow = models.Color.objects.create(name="Yellow")

    models.Fruit.objects.create(name="Apple", color=red)
    models.Fruit.objects.create(name="Banana", color=yellow)
    models.Fruit.objects.create(name="Strawberry", color=red)

    schema = strawberry.Schema(query=Query, extensions=[DjangoOptimizerExtension])

    query = """\
    query GetColors ($pagination: OffsetPaginationInput) {
      colors {
        totalCount
        results {
          fruits (pagination: $pagination) {
            totalCount
            results {
              name
            }
          }
        }
      }
    }
    """

    res = schema.execute_sync(query)
    assert res.errors is None
    assert res.data == {
        "colors": {
            "totalCount": 2,
            "results": [
                {
                    "fruits": {
                        "totalCount": 2,
                        "results": [{"name": "Apple"}, {"name": "Strawberry"}],
                    }
                },
                {
                    "fruits": {
                        "totalCount": 1,
                        "results": [{"name": "Banana"}],
                    }
                },
            ],
        }
    }


@pytest.mark.django_db(transaction=True)
def test_pagination_nested_query():
    @strawberry_django.type(models.Fruit)
    class Fruit:
        id: int
        name: str

    @strawberry_django.type(models.Color)
    class Color:
        id: int
        name: str
        fruits: OffsetPaginated[Fruit] = strawberry_django.offset_paginated()

    @strawberry.type
    class Query:
        colors: OffsetPaginated[Color] = strawberry_django.offset_paginated()

    red = models.Color.objects.create(name="Red")
    yellow = models.Color.objects.create(name="Yellow")

    models.Fruit.objects.create(name="Apple", color=red)
    models.Fruit.objects.create(name="Banana", color=yellow)
    models.Fruit.objects.create(name="Strawberry", color=red)

    schema = strawberry.Schema(query=Query)

    query = """\
    query GetColors ($pagination: OffsetPaginationInput) {
      colors {
        totalCount
        results {
          fruits (pagination: $pagination) {
            totalCount
            results {
              name
            }
          }
        }
      }
    }
    """

    res = schema.execute_sync(query)
    assert res.errors is None
    assert res.data == {
        "colors": {
            "totalCount": 2,
            "results": [
                {
                    "fruits": {
                        "totalCount": 2,
                        "results": [{"name": "Apple"}, {"name": "Strawberry"}],
                    }
                },
                {
                    "fruits": {
                        "totalCount": 1,
                        "results": [{"name": "Banana"}],
                    }
                },
            ],
        }
    }

    res = schema.execute_sync(query, variable_values={"pagination": {"limit": 1}})
    assert res.errors is None
    assert res.data == {
        "colors": {
            "totalCount": 2,
            "results": [
                {
                    "fruits": {
                        "totalCount": 2,
                        "results": [{"name": "Apple"}],
                    }
                },
                {
                    "fruits": {
                        "totalCount": 1,
                        "results": [{"name": "Banana"}],
                    }
                },
            ],
        }
    }

    res = schema.execute_sync(
        query, variable_values={"pagination": {"limit": 1, "offset": 1}}
    )
    assert res.errors is None
    assert res.data == {
        "colors": {
            "totalCount": 2,
            "results": [
                {
                    "fruits": {
                        "totalCount": 2,
                        "results": [{"name": "Strawberry"}],
                    }
                },
                {
                    "fruits": {
                        "totalCount": 1,
                        "results": [],
                    }
                },
            ],
        }
    }


@pytest.mark.django_db(transaction=True)
async def test_pagination_nested_query_async():
    @strawberry_django.type(models.Fruit)
    class Fruit:
        id: int
        name: str

    @strawberry_django.type(models.Color)
    class Color:
        id: int
        name: str
        fruits: OffsetPaginated[Fruit] = strawberry_django.offset_paginated()

    @strawberry.type
    class Query:
        colors: OffsetPaginated[Color] = strawberry_django.offset_paginated()

    red = await models.Color.objects.acreate(name="Red")
    yellow = await models.Color.objects.acreate(name="Yellow")

    await models.Fruit.objects.acreate(name="Apple", color=red)
    await models.Fruit.objects.acreate(name="Banana", color=yellow)
    await models.Fruit.objects.acreate(name="Strawberry", color=red)

    schema = strawberry.Schema(query=Query)

    query = """\
    query GetColors ($pagination: OffsetPaginationInput) {
      colors {
        totalCount
        results {
          fruits (pagination: $pagination) {
            totalCount
            results {
              name
            }
          }
        }
      }
    }
    """

    res = await schema.execute(query)
    assert res.errors is None
    assert res.data == {
        "colors": {
            "totalCount": 2,
            "results": [
                {
                    "fruits": {
                        "totalCount": 2,
                        "results": [{"name": "Apple"}, {"name": "Strawberry"}],
                    }
                },
                {
                    "fruits": {
                        "totalCount": 1,
                        "results": [{"name": "Banana"}],
                    }
                },
            ],
        }
    }

    res = await schema.execute(query, variable_values={"pagination": {"limit": 1}})
    assert res.errors is None
    assert res.data == {
        "colors": {
            "totalCount": 2,
            "results": [
                {
                    "fruits": {
                        "totalCount": 2,
                        "results": [{"name": "Apple"}],
                    }
                },
                {
                    "fruits": {
                        "totalCount": 1,
                        "results": [{"name": "Banana"}],
                    }
                },
            ],
        }
    }

    res = await schema.execute(
        query, variable_values={"pagination": {"limit": 1, "offset": 1}}
    )
    assert res.errors is None
    assert res.data == {
        "colors": {
            "totalCount": 2,
            "results": [
                {
                    "fruits": {
                        "totalCount": 2,
                        "results": [{"name": "Strawberry"}],
                    }
                },
                {
                    "fruits": {
                        "totalCount": 1,
                        "results": [],
                    }
                },
            ],
        }
    }


@pytest.mark.django_db(transaction=True)
def test_pagination_query_with_subclass():
    @strawberry_django.type(models.Fruit)
    class Fruit:
        id: int
        name: str

    @strawberry.type
    class FruitPaginated(OffsetPaginated[Fruit]):
        _custom_field: strawberry.Private[str]

        @strawberry_django.field
        def custom_field(self) -> str:
            return self._custom_field

        @classmethod
        def resolve_paginated(cls, queryset, *, info, pagination=None, **kwargs):
            return cls(
                queryset=queryset,
                pagination=pagination or OffsetPaginationInput(),
                _custom_field="pagination rocks",
            )

    @strawberry.type
    class Query:
        fruits: FruitPaginated = strawberry_django.offset_paginated()

    models.Fruit.objects.create(name="Apple")
    models.Fruit.objects.create(name="Banana")
    models.Fruit.objects.create(name="Strawberry")

    schema = strawberry.Schema(query=Query)

    query = """\
    query GetFruits ($pagination: OffsetPaginationInput) {
      fruits (pagination: $pagination) {
        totalCount
        customField
        results {
          name
        }
      }
    }
    """

    res = schema.execute_sync(query)
    assert res.errors is None
    assert res.data == {
        "fruits": {
            "totalCount": 3,
            "customField": "pagination rocks",
            "results": [{"name": "Apple"}, {"name": "Banana"}, {"name": "Strawberry"}],
        }
    }

    res = schema.execute_sync(query, variable_values={"pagination": {"limit": 1}})
    assert res.errors is None
    assert res.data == {
        "fruits": {
            "totalCount": 3,
            "customField": "pagination rocks",
            "results": [{"name": "Apple"}],
        }
    }

    res = schema.execute_sync(
        query, variable_values={"pagination": {"limit": 1, "offset": 2}}
    )
    assert res.errors is None
    assert res.data == {
        "fruits": {
            "totalCount": 3,
            "customField": "pagination rocks",
            "results": [{"name": "Strawberry"}],
        }
    }


@pytest.mark.django_db(transaction=True)
def test_pagination_query_with_resolver_schema():
    @strawberry_django.type(models.Fruit)
    class Fruit:
        id: int
        name: str

    @strawberry_django.filter_type(models.Fruit)
    class FruitFilter:
        name: str

    @strawberry_django.order(models.Fruit)
    class FruitOrder:
        name: str

    @strawberry.type
    class Query:
        @strawberry_django.offset_paginated(OffsetPaginated[Fruit])
        def fruits(self) -> QuerySet[models.Fruit]: ...

        @strawberry_django.offset_paginated(
            OffsetPaginated[Fruit],
            filters=FruitFilter,
            order=FruitOrder,
        )
        def fruits_with_order_and_filter(self) -> QuerySet[models.Fruit]: ...

    schema = strawberry.Schema(query=Query)

    expected = '''
    type Fruit {
      id: Int!
      name: String!
    }

    input FruitFilter {
      name: String!
      AND: FruitFilter
      OR: FruitFilter
      NOT: FruitFilter
      DISTINCT: Boolean
    }

    type FruitOffsetPaginated {
      pageInfo: OffsetPaginationInfo!

      """Total count of existing results."""
      totalCount: Int!

      """List of paginated results."""
      results: [Fruit!]!
    }

    input FruitOrder {
      name: String
    }

    type OffsetPaginationInfo {
      offset: Int!
      limit: Int
    }

    input OffsetPaginationInput {
      offset: Int! = 0
      limit: Int
    }

    type Query {
      fruits(pagination: OffsetPaginationInput): FruitOffsetPaginated!
      fruitsWithOrderAndFilter(filters: FruitFilter, order: FruitOrder, pagination: OffsetPaginationInput): FruitOffsetPaginated!
    }
    '''

    assert textwrap.dedent(str(schema)) == textwrap.dedent(expected).strip()


@pytest.mark.django_db(transaction=True)
def test_pagination_query_with_resolver():
    @strawberry_django.type(models.Fruit)
    class Fruit:
        id: int
        name: str

    @strawberry_django.filter_type(models.Fruit)
    class FruitFilter:
        name: strawberry.auto

    @strawberry_django.order(models.Fruit)
    class FruitOrder:
        name: strawberry.auto

    @strawberry.type
    class Query:
        @strawberry_django.offset_paginated(OffsetPaginated[Fruit])
        def fruits(self) -> QuerySet[models.Fruit]:
            return models.Fruit.objects.filter(name__startswith="S")

        @strawberry_django.offset_paginated(
            OffsetPaginated[Fruit],
            filters=FruitFilter,
            order=FruitOrder,
        )
        def fruits_with_order_and_filter(self) -> QuerySet[models.Fruit]:
            return models.Fruit.objects.filter(name__startswith="S")

    models.Fruit.objects.create(name="Apple")
    models.Fruit.objects.create(name="Strawberry")
    models.Fruit.objects.create(name="Banana")
    models.Fruit.objects.create(name="Sugar Apple")
    models.Fruit.objects.create(name="Starfruit")

    schema = strawberry.Schema(query=Query)

    query = """\
    query GetFruits (
      $pagination: OffsetPaginationInput
      $filters: FruitFilter
      $order: FruitOrder
    ) {
      fruits (pagination: $pagination) {
        totalCount
        results {
          name
        }
      }
      fruitsWithOrderAndFilter (
        pagination: $pagination
        filters: $filters
        order: $order
      ) {
        totalCount
        results {
          name
        }
      }
    }
    """

    res = schema.execute_sync(query)
    assert res.errors is None
    assert res.data == {
        "fruits": {
            "totalCount": 3,
            "results": [
                {"name": "Strawberry"},
                {"name": "Sugar Apple"},
                {"name": "Starfruit"},
            ],
        },
        "fruitsWithOrderAndFilter": {
            "totalCount": 3,
            "results": [
                {"name": "Strawberry"},
                {"name": "Sugar Apple"},
                {"name": "Starfruit"},
            ],
        },
    }

    res = schema.execute_sync(query, variable_values={"pagination": {"limit": 1}})
    assert res.errors is None
    assert res.data == {
        "fruits": {
            "totalCount": 3,
            "results": [
                {"name": "Strawberry"},
            ],
        },
        "fruitsWithOrderAndFilter": {
            "totalCount": 3,
            "results": [
                {"name": "Strawberry"},
            ],
        },
    }

    res = schema.execute_sync(
        query,
        variable_values={
            "pagination": {"limit": 2},
            "order": {"name": "ASC"},
            "filters": {"name": "Strawberry"},
        },
    )
    assert res.errors is None
    assert res.data == {
        "fruits": {
            "totalCount": 3,
            "results": [
                {"name": "Strawberry"},
                {"name": "Sugar Apple"},
            ],
        },
        "fruitsWithOrderAndFilter": {
            "totalCount": 1,
            "results": [
                {"name": "Strawberry"},
            ],
        },
    }


@pytest.mark.django_db(transaction=True)
def test_pagination_query_with_resolver_arguments():
    @strawberry_django.type(models.Fruit)
    class Fruit:
        id: int
        name: str

    @strawberry_django.filter_type(models.Fruit)
    class FruitFilter:
        name: strawberry.auto

    @strawberry_django.order(models.Fruit)
    class FruitOrder:
        name: strawberry.auto

    @strawberry.type
    class Query:
        @strawberry_django.offset_paginated(OffsetPaginated[Fruit])
        def fruits(self, starts_with: str) -> QuerySet[models.Fruit]:
            return models.Fruit.objects.filter(name__startswith=starts_with)

        @strawberry_django.offset_paginated(
            OffsetPaginated[Fruit],
            filters=FruitFilter,
            order=FruitOrder,
        )
        def fruits_with_order_and_filter(
            self, starts_with: str
        ) -> QuerySet[models.Fruit]:
            return models.Fruit.objects.filter(name__startswith=starts_with)

    models.Fruit.objects.create(name="Apple")
    models.Fruit.objects.create(name="Strawberry")
    models.Fruit.objects.create(name="Banana")
    models.Fruit.objects.create(name="Sugar Apple")
    models.Fruit.objects.create(name="Starfruit")

    schema = strawberry.Schema(query=Query)

    query = """\
    query GetFruits (
      $pagination: OffsetPaginationInput
      $filters: FruitFilter
      $order: FruitOrder
      $startsWith: String!
    ) {
      fruits (startsWith: $startsWith, pagination: $pagination) {
        totalCount
        results {
          name
        }
      }
      fruitsWithOrderAndFilter (
        startsWith: $startsWith
        pagination: $pagination
        filters: $filters
        order: $order
      ) {
        totalCount
        results {
          name
        }
      }
    }
    """

    res = schema.execute_sync(query, variable_values={"startsWith": "S"})
    assert res.errors is None
    assert res.data == {
        "fruits": {
            "totalCount": 3,
            "results": [
                {"name": "Strawberry"},
                {"name": "Sugar Apple"},
                {"name": "Starfruit"},
            ],
        },
        "fruitsWithOrderAndFilter": {
            "totalCount": 3,
            "results": [
                {"name": "Strawberry"},
                {"name": "Sugar Apple"},
                {"name": "Starfruit"},
            ],
        },
    }

    res = schema.execute_sync(
        query,
        variable_values={"startsWith": "S", "pagination": {"limit": 1}},
    )
    assert res.errors is None
    assert res.data == {
        "fruits": {
            "totalCount": 3,
            "results": [
                {"name": "Strawberry"},
            ],
        },
        "fruitsWithOrderAndFilter": {
            "totalCount": 3,
            "results": [
                {"name": "Strawberry"},
            ],
        },
    }

    res = schema.execute_sync(
        query,
        variable_values={
            "startsWith": "S",
            "pagination": {"limit": 2},
            "order": {"name": "ASC"},
            "filters": {"name": "Strawberry"},
        },
    )
    assert res.errors is None
    assert res.data == {
        "fruits": {
            "totalCount": 3,
            "results": [
                {"name": "Strawberry"},
                {"name": "Sugar Apple"},
            ],
        },
        "fruitsWithOrderAndFilter": {
            "totalCount": 1,
            "results": [
                {"name": "Strawberry"},
            ],
        },
    }


@pytest.mark.django_db(transaction=True)
@override_settings(
    STRAWBERRY_DJANGO=StrawberryDjangoSettings(  # type: ignore
        PAGINATION_DEFAULT_LIMIT=2,
    ),
)
def test_pagination_default_limit():
    @strawberry_django.type(models.Fruit)
    class Fruit:
        id: int
        name: str

    @strawberry.type
    class Query:
        fruits: OffsetPaginated[Fruit] = strawberry_django.offset_paginated()

    models.Fruit.objects.create(name="Apple")
    models.Fruit.objects.create(name="Banana")
    models.Fruit.objects.create(name="Strawberry")
    models.Fruit.objects.create(name="Watermelon")

    schema = strawberry.Schema(query=Query)

    query = """\
    query GetFruits ($pagination: OffsetPaginationInput) {
      fruits (pagination: $pagination) {
        totalCount
        results {
          name
        }
      }
    }
    """

    res = schema.execute_sync(query)
    assert res.errors is None
    assert res.data == {
        "fruits": {
            "totalCount": 4,
            "results": [{"name": "Apple"}, {"name": "Banana"}],
        }
    }

    res = schema.execute_sync(query, variable_values={"pagination": {"offset": 1}})
    assert res.errors is None
    assert res.data == {
        "fruits": {
            "totalCount": 4,
            "results": [{"name": "Banana"}, {"name": "Strawberry"}],
        }
    }

    # Setting limit to None should use default limit (same as not providing limit)
    res = schema.execute_sync(query, variable_values={"pagination": {"limit": None}})
    assert res.errors is None
    assert res.data == {
        "fruits": {
            "totalCount": 4,
            "results": [
                {"name": "Apple"},
                {"name": "Banana"},
            ],
        }
    }


@pytest.mark.django_db(transaction=True)
@pytest.mark.parametrize(
    ("requested_limit", "max_limit", "default_limit", "expected_limit"),
    [
        (5, None, 100, 5),  # Explicit limit should be honored
        (None, 50, 20, 20),  # None should use default_limit
        (None, 50, None, 50),  # None with no default should use max_limit
        (None, None, 20, 20),  # None with no max should use default_limit
        (100, 50, 20, 50),  # Limit > max should be capped to max_limit
        ("UNSET", 50, 20, 20),  # UNSET should use default_limit
        ("UNSET", None, 20, 20),  # UNSET with no max should use default_limit
        (10, 50, 5, 10),  # Explicit limit < max should be honored
        # Edge cases
        (
            None,
            None,
            None,
            100,
        ),  # None with no settings should use global default (100)
        ("UNSET", 50, None, 50),  # UNSET with no default should fall back to max_limit
        (
            "UNSET",
            None,
            None,
            100,
        ),  # UNSET with no settings should use global default (100)
        (-5, 50, 20, 50),  # Negative limit should clamp to max_limit when set
        (
            -5,
            None,
            20,
            100,
        ),  # Negative limit without explicit max clamps to the default max (100)
        (
            -5,
            None,
            None,
            100,
        ),  # Negative limit without any settings clamps to the default max (100)
        # Large positive limits should be clamped to max_limit
        (
            9999999,
            None,
            None,
            100,
        ),  # Very large limit with default settings → clamped to default max (100)
        (
            9999999,
            500,
            20,
            500,
        ),  # Very large limit with overridden max → clamped to max_limit
    ],
)
def test_page_info_reflects_effective_limit(
    requested_limit, max_limit, default_limit, expected_limit
):
    """Test that pageInfo.limit reflects the actual limit applied, not the requested one."""

    @strawberry_django.type(models.Fruit)
    class Fruit:
        id: int
        name: str

    @strawberry.type
    class Query:
        fruits: OffsetPaginated[Fruit] = strawberry_django.offset_paginated()

    for i in range(10):
        models.Fruit.objects.create(name=f"Fruit{i}")

    schema = strawberry.Schema(query=Query)

    query = """\
    query GetFruits ($pagination: OffsetPaginationInput) {
      fruits (pagination: $pagination) {
        pageInfo {
          limit
          offset
        }
        totalCount
        results {
          name
        }
      }
    }
    """

    settings_dict = {}
    if max_limit is not None:
        settings_dict["PAGINATION_MAX_LIMIT"] = max_limit
    if default_limit is not None:
        settings_dict["PAGINATION_DEFAULT_LIMIT"] = default_limit

    with override_settings(STRAWBERRY_DJANGO=settings_dict):
        if requested_limit == "UNSET":
            # Don't provide pagination at all
            res = schema.execute_sync(query)
        elif requested_limit is None:
            # Explicitly pass null
            res = schema.execute_sync(
                query, variable_values={"pagination": {"limit": None}}
            )
        else:
            # Pass explicit limit
            res = schema.execute_sync(
                query, variable_values={"pagination": {"limit": requested_limit}}
            )

    assert res.errors is None
    assert res.data is not None
    assert res.data["fruits"]["pageInfo"]["limit"] == expected_limit
    assert res.data["fruits"]["pageInfo"]["offset"] == 0

    results = res.data["fruits"]["results"]
    total_fruits = 10
    if expected_limit is not None and expected_limit > 0:
        expected_count = min(expected_limit, total_fruits)
    else:
        expected_count = total_fruits
    assert len(results) == expected_count


@pytest.mark.django_db(transaction=True)
def test_offset_paginated_extensions_receive_filters():
    @strawberry_django.filter_type(models.Fruit, lookups=True)
    class FruitFilter:
        name: strawberry.auto

    @strawberry_django.type(models.Fruit, filters=FruitFilter)
    class Fruit:
        id: int
        name: str

    class CaptureKwargsExtension(FieldExtension):
        def __init__(self):
            super().__init__()
            self.seen_kwargs: dict[str, object] | None = None

        def resolve(self, next_, source, info, **kwargs):
            self.seen_kwargs = dict(kwargs)
            return next_(source, info, **kwargs)

    extension = CaptureKwargsExtension()

    @strawberry.type
    class Query:
        @strawberry_django.offset_paginated(
            OffsetPaginated[Fruit],
            filters=FruitFilter,
            extensions=[extension],
        )
        def fruits(self) -> QuerySet[models.Fruit]:
            return models.Fruit.objects.all()

    schema = strawberry.Schema(query=Query)
    models.Fruit.objects.create(name="Apple")

    query = """
      query ($filters: FruitFilter) {
        fruits(filters: $filters) {
          results {
            name
          }
        }
      }
    """

    res = schema.execute_sync(
        query,
        variable_values={"filters": {"name": {"exact": "Apple"}}},
    )

    assert res.errors is None
    assert extension.seen_kwargs is not None
    assert "filters" in extension.seen_kwargs


@pytest.mark.django_db(transaction=True)
def test_offset_paginated_allows_filters_without_resolver_param():
    @strawberry_django.filter_type(models.Fruit, lookups=True)
    class FruitFilter:
        name: strawberry.auto

    @strawberry_django.type(models.Fruit, filters=FruitFilter)
    class Fruit:
        id: int
        name: str

    @strawberry.type
    class Query:
        @strawberry_django.offset_paginated(
            OffsetPaginated[Fruit],
            filters=FruitFilter,
        )
        def fruits(self) -> QuerySet[models.Fruit]:
            return models.Fruit.objects.all()

    schema = strawberry.Schema(query=Query)
    models.Fruit.objects.create(name="Apple")
    models.Fruit.objects.create(name="Banana")

    query = """
      query ($filters: FruitFilter) {
        fruits(filters: $filters) {
          results {
            name
          }
        }
      }
    """

    res = schema.execute_sync(
        query,
        variable_values={"filters": {"name": {"exact": "Apple"}}},
    )

    assert res.errors is None
    assert res.data == {"fruits": {"results": [{"name": "Apple"}]}}



import pathlib

import strawberry
from pytest_snapshot.plugin import Snapshot

import strawberry_django
from strawberry_django import mutations
from tests.conftest import normalize_sdl

from .models import Issue, Milestone, Project
from .schema import IssueInput, IssueType, MilestoneType, ProjectType, schema

SNAPSHOTS_DIR = pathlib.Path(__file__).parent / "snapshots"


def test_schema(snapshot: Snapshot):
    snapshot.snapshot_dir = SNAPSHOTS_DIR
    snapshot.assert_match(normalize_sdl(str(schema)), "schema.gql")


def test_schema_with_inheritance(snapshot: Snapshot):
    @strawberry_django.type(Project)
    class ProjectTypeSubclass(ProjectType): ...

    @strawberry_django.type(Milestone)
    class MilestoneTypeSubclass(MilestoneType): ...

    @strawberry_django.input(Issue)
    class IssueInputSubclass(IssueInput): ...

    @strawberry.type
    class Query:
        project: ProjectTypeSubclass | None = strawberry_django.node()
        milestone: MilestoneTypeSubclass | None = strawberry_django.node()

    @strawberry.type
    class Mutation:
        create_issue: IssueType = mutations.create(
            IssueInputSubclass,
            handle_django_errors=True,
            argument_name="input",
        )

    schema = strawberry.Schema(query=Query, mutation=Mutation)
    snapshot.snapshot_dir = SNAPSHOTS_DIR
    snapshot.assert_match(normalize_sdl(str(schema)), "schema_with_inheritance.gql")



from typing import cast

import pytest
import strawberry
from strawberry import auto
from strawberry.types import get_object_definition
from strawberry.types.base import StrawberryOptional

import strawberry_django
from strawberry_django.filters import get_django_model_filter_input_type
from strawberry_django.settings import strawberry_django_settings
from tests import models

DjangoModelFilterInput = get_django_model_filter_input_type()


@pytest.fixture(autouse=True)
def _filter_order_settings(settings):
    settings.STRAWBERRY_DJANGO = {
        **strawberry_django_settings(),
        "USE_DEPRECATED_FILTERS": True,
    }


def test_filter():
    @strawberry_django.filters.filter_type(models.Fruit)
    class Filter:
        id: auto
        name: auto
        color: auto
        types: auto

    object_definition = get_object_definition(Filter, strict=True)
    assert [(f.name, f.type) for f in object_definition.fields] == [
        ("id", StrawberryOptional(cast("type", strawberry.ID))),
        ("name", StrawberryOptional(str)),
        ("color", StrawberryOptional(DjangoModelFilterInput)),
        ("types", StrawberryOptional(DjangoModelFilterInput)),
        ("AND", StrawberryOptional(Filter)),
        ("OR", StrawberryOptional(Filter)),
        ("NOT", StrawberryOptional(Filter)),
        ("DISTINCT", StrawberryOptional(bool)),
    ]


def test_lookups():
    @strawberry_django.filters.filter_type(models.Fruit, lookups=True)
    class Filter:
        id: auto
        name: auto
        color: auto
        types: auto

    object_definition = get_object_definition(Filter, strict=True)
    assert [
        (f.name, f.type.of_type.__name__)  # type: ignore
        for f in object_definition.fields
    ] == [
        ("id", "FilterLookup"),
        ("name", "FilterLookup"),
        ("color", "DjangoModelFilterInput"),
        ("types", "DjangoModelFilterInput"),
        ("AND", "Filter"),
        ("OR", "Filter"),
        ("NOT", "Filter"),
        ("DISTINCT", "bool"),
    ]


def test_inherit(testtype):
    @testtype(models.Fruit)
    class Base:
        id: auto
        name: auto
        color: auto
        types: auto

    @strawberry_django.filters.filter_type(models.Fruit)
    class Filter(Base):
        pass

    object_definition = get_object_definition(Filter, strict=True)
    assert [(f.name, f.type) for f in object_definition.fields] == [
        ("id", StrawberryOptional(cast("type", strawberry.ID))),
        ("name", StrawberryOptional(str)),
        ("color", StrawberryOptional(DjangoModelFilterInput)),
        ("types", StrawberryOptional(DjangoModelFilterInput)),
        ("AND", StrawberryOptional(Filter)),
        ("OR", StrawberryOptional(Filter)),
        ("NOT", StrawberryOptional(Filter)),
        ("DISTINCT", StrawberryOptional(bool)),
    ]


def test_relationship():
    @strawberry_django.filters.filter_type(models.Color)
    class ColorFilter:
        name: auto

    @strawberry_django.filters.filter_type(models.Fruit)
    class Filter:
        color: ColorFilter | None

    object_definition = get_object_definition(Filter, strict=True)
    assert [(f.name, f.type) for f in object_definition.fields] == [
        ("color", StrawberryOptional(ColorFilter)),
        ("AND", StrawberryOptional(Filter)),
        ("OR", StrawberryOptional(Filter)),
        ("NOT", StrawberryOptional(Filter)),
        ("DISTINCT", StrawberryOptional(bool)),
    ]


def test_relationship_with_inheritance():
    @strawberry_django.filters.filter_type(models.Color)
    class ColorFilter:
        name: auto

    @strawberry_django.type(models.Fruit)
    class Base:
        color: auto

    @strawberry_django.filters.filter_type(models.Fruit)
    class Filter(Base):
        color: ColorFilter | None

    object_definition = get_object_definition(Filter, strict=True)
    assert [(f.name, f.type) for f in object_definition.fields] == [
        ("color", StrawberryOptional(ColorFilter)),
        ("AND", StrawberryOptional(Filter)),
        ("OR", StrawberryOptional(Filter)),
        ("NOT", StrawberryOptional(Filter)),
        ("DISTINCT", StrawberryOptional(bool)),
    ]



import pytest
from django.test.client import Client

from strawberry_django.test.client import AsyncTestClient
from tests.utils import GraphQLTestClient

query_to_non_existed_field = "{ nonExistentField { id } }"


def check_non_existed_field_error(errors):
    assert isinstance(errors, list)
    assert len(errors) == 1
    error = errors[0]
    assert isinstance(error, dict)
    assert "nonExistentField" in error["message"]
    assert "Cannot query field" in error["message"]
    assert error["locations"]


def test_client_assert_no_errors_verbose_message(db):
    """Test that GraphQLTestClient (sync) raises AssertionError with verbose error messages.

    This test directly uses GraphQLTestClient with a sync Client to verify the verbose
    error message functionality in the sync client implementation.
    """
    client = GraphQLTestClient("/graphql/", Client())

    with pytest.raises(AssertionError) as exc_info:
        client.query(query_to_non_existed_field)

    check_non_existed_field_error(exc_info.value.args[0])


@pytest.mark.asyncio
async def test_async_client_assert_no_errors_verbose_message(db):
    """Test that AsyncTestClient raises AssertionError with verbose error messages.

    This test directly uses AsyncTestClient to verify the verbose error message
    functionality in the async client implementation.
    """
    client = AsyncTestClient("/graphql_async/")

    with pytest.raises(AssertionError) as exc_info:
        await client.query(query_to_non_existed_field)

    check_non_existed_field_error(exc_info.value.args[0])
